from enum import auto

from mstrio.utils import helper
from mstrio.utils.enum_helper import AutoName, get_enum_val
from mstrio.utils.enums import DaysOfWeek as _DaysOfWeek
from mstrio.utils.helper import Dictable


class ScheduleEnums:
    """
    Object representations of recurrence information of a time-based schedule
    """

    DaysOfWeek = _DaysOfWeek

    class WeekOffset(AutoName):
        FIRST = auto()
        SECOND = auto()
        THIRD = auto()
        FOURTH = auto()
        LAST = auto()
        NONE = None

    class RecurrencePattern(AutoName):
        DAILY = auto()
        WEEKLY = auto()
        MONTHLY = auto()
        YEARLY = auto()

    class ExecutionPattern(AutoName):
        ONCE = auto()
        REPEAT = auto()

    class DailyPattern(AutoName):
        DAY = auto()
        WEEKDAY = auto()
        NONE = None

    class MonthlyPattern(AutoName):
        DAY = auto()
        DAY_OF_WEEK = auto()
        WEEKDAY = auto()
        LAST_DAY = auto()
        DAYS_OF_MONTH = auto()
        NONE = None

    class WeekdayOffset(AutoName):
        FIRST = auto()
        LAST = auto()
        NONE = None

    class YearlyPattern(AutoName):
        DAY = auto()
        DAY_OF_WEEK = auto()
        NONE = None


class ScheduleTime(Dictable):
    """Object representation of details of time-based schedule"""

    class Execution(Dictable):
        """Object representation of the execution information of the schedule"""

        _FROM_DICT_MAP = {
            'execution_pattern': ScheduleEnums.ExecutionPattern,
        }

        def __init__(
            self,
            execution_pattern: (
                ScheduleEnums.ExecutionPattern | str
            ) = ScheduleEnums.ExecutionPattern.ONCE,
            execution_time: str | None = None,
            start_time: str | None = None,
            stop_time: str | None = None,
            repeat_interval: int | None = 1,
        ):
            """Set attributes representing execution information of the schedule

            Args:
                execution_pattern (ScheduleEnums.ExecutionPattern, optional):
                    The execution pattern of the schedule.
                    Possible values are ONCE or REPEAT.
                    Defaults to ScheduleEnums.ExecutionPattern.ONCE.
                execution_time (str, optional):
                    The execution time of the execution day, if
                    execution_pattern is once. Format should be HH:mm:ss.
                    Defaults to ''.
                start_time (str, optional):
                    The start time of the execution day, if execution_pattern
                    is repeat. Format should be HH:mm:ss. Defaults to ''.
                stop_time (str, optional):
                    The stop time of the execution day, if execution_pattern is
                    repeat. Format should be HH:mm:ss. Defaults to ''.
                repeat_interval (int, optional):
                    The repeat interval of minutes of the execution day, if
                    executionPattern is repeat. Defaults to 1.
            Raises:
                ValueError: when execution_repeat_interval = None if
                execution_pattern == REPEAT
            """

            if execution_pattern == ScheduleEnums.ExecutionPattern.REPEAT:
                if repeat_interval is None:
                    msg = (
                        'Value error: `execution_repeat_interval` cannot be `None`, '
                        'for `execution_pattern == ExecutionPattern.REPEAT`'
                    )
                    helper.exception_handler(
                        msg=msg,
                        exception_type=ValueError,
                    )
                self.start_time = start_time
                self.stop_time = stop_time
                self.repeat_interval = repeat_interval
            else:
                self.execution_time = execution_time

            self.execution_pattern = (
                execution_pattern
                if isinstance(execution_pattern, ScheduleEnums.ExecutionPattern)
                else ScheduleEnums.ExecutionPattern(execution_pattern)
            )

    class Daily(Dictable):
        """
        Object representation of daily recurrence information of the schedule
        """

        _FROM_DICT_MAP = {'daily_pattern': ScheduleEnums.DailyPattern}

        def __init__(
            self,
            daily_pattern: (
                ScheduleEnums.DailyPattern | str
            ) = ScheduleEnums.DailyPattern.DAY,
            repeat_interval: int = 1,
        ):
            """Set attributes representing daily recurrence information of the
            schedule

            Args:
                daily_pattern (ScheduleEnums.DailyPattern, optional):
                    The daily recurrence pattern of the schedule.Possible values
                    are DAY, WEEKDAY.Defaults to ScheduleEnums.DailyPattern.DAY.
                repeat_interval (int, optional):
                    The repeat interval of days of daily schedule,
                    if daily_pattern is DAY. Defaults to 1.
            """
            self.daily_pattern = (
                daily_pattern
                if isinstance(daily_pattern, ScheduleEnums.DailyPattern)
                else ScheduleEnums.DailyPattern(daily_pattern)
            )
            if daily_pattern == ScheduleEnums.DailyPattern.DAY:
                self.repeat_interval = repeat_interval

    class Weekly(Dictable):
        """Object representation of weekly recurrence information of the
        schedule"""

        _FROM_DICT_MAP = {'days_of_week': [ScheduleEnums.DaysOfWeek]}

        def __init__(
            self,
            days_of_week: list[ScheduleEnums.DaysOfWeek] | list[str] | None = None,
            repeat_interval: int = 1,
        ):
            """Set attributes representing weekly recurrence information of the
            schedule.

            Args:
                days_of_week (List[ScheduleEnums.DaysOfWeek], optional):
                    The days of week of weekly schedule. Possible values are:
                    MONDAY, TUESDAY, WEDNESDAY, THURSDAY, FRIDAY, SATURDAY,
                    SUNDAY. Defaults to None.
                repeat_interval (int, optional):
                    The repeat interval of weeks of weekly schedule.
                    Defaults to 1.
            """
            self.days_of_week = [
                (
                    ScheduleEnums.DaysOfWeek(item)
                    if not isinstance(item, ScheduleEnums.DaysOfWeek)
                    else item
                )
                for item in (days_of_week or [])
            ]

            self.repeat_interval = repeat_interval

    class Monthly(Dictable):
        """Object representation of monthly recurrence information of the
        schedule"""

        _FROM_DICT_MAP = {
            'monthly_pattern': ScheduleEnums.MonthlyPattern,
            'week_offset': ScheduleEnums.WeekOffset,
            'day_of_week': ScheduleEnums.DaysOfWeek,
            'weekday_offset': ScheduleEnums.WeekdayOffset,
        }

        def __init__(
            self,
            monthly_pattern: ScheduleEnums.MonthlyPattern | str,
            repeat_interval: int,
            day: int | None = None,
            week_offset: ScheduleEnums.WeekOffset | str | None = None,
            day_of_week: ScheduleEnums.DaysOfWeek | str | None = None,
            weekday_offset: ScheduleEnums.WeekdayOffset | str | None = None,
            days_of_month: str | None = None,
        ):
            """Set attributes representing monthly recurrence information of the
            schedule.

            Args:
                monthly_pattern (ScheduleEnums.MonthlyPattern):
                    The monthly recurrence pattern of the schedule. Possible
                    values are: DAY, DAY_OF_WEEK, WEEKDAY, LAST_DAY,
                    DAYS_OF_MONTH.
                repeat_interval (int):
                    The repeat interval of months of monthly schedule.
                day (Optional[int], optional):
                    The day in month of monthly schedule, if monthly_pattern is
                    DAY. Defaults to None.
                week_offset (Optional[ScheduleEnums.WeekOffset], optional):
                    The week offset in month of monthly schedule, if
                    monthly_pattern is DAY_OF_WEEK. Possible values are:
                    FIRST, SECOND, THIRD, FOURTH, LAST. Defaults to None.
                day_of_week (Optional[ScheduleEnums.DaysOfWeek], optional):
                    The day of week in month of monthly schedule, if
                    monthly_pattern is DAY_OF_WEEK. Possible values are:
                    MONDAY, TUESDAY, WEDNESDAY, THURSDAY, FRIDAY, SATURDAY,
                    SUNDAY. Defaults to None.
                weekday_offset (Optional[ScheduleEnums.WeekdayOffset], optional)
                    The weekday offset in month of monthly schedule,
                    if monthly_pattern is WEEKDAY. Possible values are: FIRST,
                    LAST. Defaults to None.
                days_of_month (Optional[str], optional):
                    The days of month of monthly schedule, if monthly_pattern
                    is DAYS_OF_MONTH. Must be provided as a list of one or more
                    stringified digits (from '1' to '31'). Defaults to None.
            """

            self.monthly_pattern = (
                monthly_pattern
                if isinstance(monthly_pattern, ScheduleEnums.MonthlyPattern)
                else ScheduleEnums.MonthlyPattern(monthly_pattern)
            )
            self.repeat_interval = repeat_interval

            if monthly_pattern == ScheduleEnums.MonthlyPattern.DAY:
                self.day = day
            elif monthly_pattern == ScheduleEnums.MonthlyPattern.DAY_OF_WEEK:
                self.week_offset = (
                    week_offset
                    if isinstance(week_offset, ScheduleEnums.WeekOffset)
                    else ScheduleEnums.WeekOffset(week_offset)
                )
                self.day_of_week = (
                    day_of_week
                    if isinstance(day_of_week, ScheduleEnums.DaysOfWeek)
                    else ScheduleEnums.DaysOfWeek(day_of_week)
                )
            elif monthly_pattern == ScheduleEnums.MonthlyPattern.WEEKDAY:
                self.weekday_offset = (
                    weekday_offset
                    if isinstance(weekday_offset, ScheduleEnums.WeekdayOffset)
                    else ScheduleEnums.WeekdayOffset(weekday_offset)
                )
            elif monthly_pattern == ScheduleEnums.MonthlyPattern.DAYS_OF_MONTH:
                self.days_of_month = days_of_month

    class Yearly(Dictable):
        """Object representation of yearly recurrence information of the
        schedule"""

        _FROM_DICT_MAP = {
            'yearly_pattern': ScheduleEnums.YearlyPattern,
            'week_offset': ScheduleEnums.WeekOffset,
            'day_of_week': ScheduleEnums.DaysOfWeek,
        }

        def __init__(
            self,
            yearly_pattern: (
                ScheduleEnums.YearlyPattern | str
            ) = ScheduleEnums.YearlyPattern.DAY,
            month: int = 1,
            day: int = 1,
            week_offset: ScheduleEnums.WeekOffset | str | None = None,
            day_of_week: ScheduleEnums.DaysOfWeek | str | None = None,
        ):
            """Set attributes representing yearly recurrence information of the
            schedule

            Args:
                yearly_pattern (ScheduleEnums.YearlyPattern, optional):
                    The yearly recurrence pattern of the schedule. Possible
                    values are DAY, DAY_OF_WEEK.
                    Defaults to ScheduleEnums.YearlyPattern.DAY.
                month (int, optional):
                    The month in year of yearly schedule. Defaults to 1.
                day (int, optional):
                    The day in month of yearly schedule, if yearly_pattern
                    is DAY. Defaults to 1.
                week_offset (ScheduleEnums.WeekOffset, optional):
                    The week offset in year of yearly schedule, if
                    yearly_pattern is DAY_OF_WEEK. Possible values are FIRST,
                    SECOND, THIRD, FOURTH, LAST. Defaults to None.
                day_of_week (ScheduleEnums.DaysOfWeek, optional):
                    The day of week in year of yearly schedule,
                    if yearly_pattern is DAY_OF_WEEK. Possible values are:
                    MONDAY, TUESDAY, WEDNESDAY, THURSDAY, FRIDAY, SATURDAY,
                    SUNDAY. Defaults to None.
            """
            self.yearly_pattern = (
                yearly_pattern
                if isinstance(yearly_pattern, ScheduleEnums.YearlyPattern)
                else ScheduleEnums.YearlyPattern(yearly_pattern)
            )
            self.month = month
            if yearly_pattern == ScheduleEnums.YearlyPattern.DAY:
                self.day = day
            if yearly_pattern == ScheduleEnums.YearlyPattern.DAY_OF_WEEK:
                self.week_offset = (
                    week_offset
                    if isinstance(week_offset, ScheduleEnums.WeekOffset)
                    else ScheduleEnums.WeekOffset(week_offset)
                )
                self.day_of_week = (
                    day_of_week
                    if isinstance(day_of_week, ScheduleEnums.DaysOfWeek)
                    else ScheduleEnums.DaysOfWeek(day_of_week)
                )

    _FROM_DICT_MAP = {
        'recurrence_pattern': ScheduleEnums.RecurrencePattern,
        'days_of_week': ScheduleEnums.DaysOfWeek,
        'execution': Execution.from_dict,
        'daily': Daily.from_dict,
        'weekly': Weekly.from_dict,
        'monthly': Monthly.from_dict,
        'yearly': Yearly.from_dict,
    }

    def __init__(
        self,
        recurrence_pattern: ScheduleEnums.RecurrencePattern | str | None = None,
        execution: Execution | None = None,
        daily: Daily | None = None,
        weekly: Weekly | None = None,
        monthly: Monthly | None = None,
        yearly: Yearly | None = None,
    ):
        """Set attributes representing details of the time-based schedule.

        Args:
            recurrence_pattern (ScheduleEnums.RecurrencePattern, optional):
                The recurrence pattern of the schedule. Possible values are
                DAILY, WEEKLY, MONTHLY, YEARLY. Defaults to None.
            execution (Execution, optional):
                Object representing the execution information of the schedule.
                Defaults to None.
            daily (Daily, optional):
                Object representing daily recurrence information of the
                schedule. Defaults to None.
            weekly (Weekly, optional):
                Object representing weekly recurrence information of the
                schedule. Defaults to None.
            monthly (Monthly, optional):
                Object representing monthly recurrence information of the
                schedule. Defaults to None.
            yearly (Yearly, optional):
                Object representing yearly recurrence information of the
                schedule. Defaults to None.
        """

        self.recurrence_pattern = (
            recurrence_pattern
            if isinstance(recurrence_pattern, ScheduleEnums.RecurrencePattern)
            else ScheduleEnums.RecurrencePattern(recurrence_pattern)
        )
        self.execution = execution
        if daily:
            self.daily = daily
        elif weekly:
            self.weekly = weekly
        elif monthly:
            self.monthly = monthly
        elif yearly:
            self.yearly = yearly

    @classmethod
    def from_details(
        cls,
        recurrence_pattern: ScheduleEnums.RecurrencePattern | str,
        execution_pattern: ScheduleEnums.ExecutionPattern | str,
        execution_time: str | None = None,
        start_time: str | None = None,
        stop_time: str | None = None,
        execution_repeat_interval: int | None = None,
        daily_pattern: (
            ScheduleEnums.DailyPattern | str | None
        ) = ScheduleEnums.DailyPattern.NONE,
        repeat_interval: int | None = None,
        days_of_week: list[ScheduleEnums.DaysOfWeek] | list[str] | None = None,
        day: int | None = None,
        month: int | None = None,
        week_offset: (
            ScheduleEnums.WeekOffset | str | None
        ) = ScheduleEnums.WeekOffset.NONE,
        day_of_week: (
            ScheduleEnums.DaysOfWeek | str | None
        ) = ScheduleEnums.DaysOfWeek.NONE,
        weekday_offset: str | None = None,
        days_of_month: list[str] | None = None,
        monthly_pattern: (
            ScheduleEnums.MonthlyPattern | str | None
        ) = ScheduleEnums.MonthlyPattern.NONE,
        yearly_pattern: (
            ScheduleEnums.YearlyPattern | str | None
        ) = ScheduleEnums.YearlyPattern.NONE,
    ):
        """
        Uses provided properties to create object representation of details of a
        time-based schedule.

        This object is used as 'time' part of request body and is needed for
        creating and updating schedule.

        Args:
            recurrence_pattern (ScheduleEnums.RecurrencePattern, optional):
                The recurrence pattern of the schedule.
                Possible values are DAILY, WEEKLY, MONTHLY, YEARLY
                Defaults to None.
            execution_pattern (ScheduleEnums.ExecutionPattern, optional):
                The execution pattern of the schedule.
                Possible values are ONCE, REPEAT. Defaults to None.
            execution_time (str, optional):
                The execution time of the execution day, if execution_pattern
                is ONCE. Format should be HH:mm:ss. Defaults to ''.
            start_time (str, optional):
                The start time of the execution day, if execution_pattern
                is REPEAT. Format should be HH:mm:ss. Defaults to ''.
            stop_time (str, optional):
                The stop time of the execution day, if execution_pattern
                is REPEAT. Format should be HH:mm:ss.
                Defaults to ''.
            execution_repeat_interval (int, optional):
                The repeat interval of minutes of the execution day, if
                execution_pattern is REPEAT. Defaults to None.
            daily_pattern (ScheduleEnums.DailyPattern, optional):
                The daily recurrence pattern of the schedule.
                Possible values are DAY, WEEKDAY.
                Defaults to None.
            repeat_interval (int, optional):
                The repeat interval of days of daily schedule, if daily_pattern
                is DAY or, The repeat interval of weeks of weekly schedule or,
                The repeat interval of months of monthly schedule.
                Defaults to None.
            day (int, optional):
                The day in month of monthly schedule, if monthly_pattern is DAY
                or, The day in month of yearly schedule, if yearly_pattern is
                DAY. Defaults to None.
            month (int, optional):
                The month in year of yearly schedule.
                Defaults to None.
            week_offset (ScheduleEnums.WeekOffset, optional):
                The week offset in month of monthly schedule, if monthly_pattern
                is DAY_OF_WEEK or, The week offset in year of yearly schedule,
                if yearly_pattern is DAY_OF_WEEK. Possible values are FIRST,
                SECOND, THIRD, FOURTH, LAST. Defaults to None.
            day_of_week (ScheduleEnums.DaysOfWeek, optional):
                The days of week of weekly schedule or, The day of week in month
                of monthly schedule, if monthly_pattern is DAY_OF_WEEK or,
                The day of week in year of yearly schedule, if yearly_pattern is
                DAY_OF_WEEK. Possible values are: MONDAY, TUESDAY, WEDNESDAY,
                THURSDAY, FRIDAY, SATURDAY, SUNDAY. Defaults to None.
            weekday_offset (ScheduleEnums.WeekdayOffset, optional):
                The weekday offset in month of monthly schedule, if
                monthly_pattern is WEEKDAY. Possible values are: FIRST, LAST.
                Defaults to None.
            days_of_month (List[str], optional):
                The days of month of monthly schedule, if monthly_pattern is
                DAYS_OF_MONTH. Must be provided as a list of one or more
                stringified digits (from '1' to '31'). Defaults to None.
            monthly_pattern (ScheduleEnums.MonthlyPattern, optional):
                The monthly recurrence pattern of the schedule. Possible values
                are: DAY, DAY_OF_WEEK, WEEKDAY, LAST_DAY, DAYS_OF_MONTH.
                Defaults to None.
            yearly_pattern (ScheduleEnums.YearlyPattern, optional):
                The yearly recurrence pattern of the schedule. Possible values
                are DAY, DAY_OF_WEEK. Defaults to None.
        Returns:
            ScheduleTime: Object representation of details of time-based
            schedule.
        """

        execution_pattern = get_enum_val(
            execution_pattern, ScheduleEnums.ExecutionPattern
        )

        if execution_pattern == ScheduleEnums.ExecutionPattern.ONCE.value:
            execution = cls.Execution.from_dict(
                {
                    'execution_pattern': execution_pattern,
                    'execution_time': execution_time,
                }
            )
        elif execution_pattern == ScheduleEnums.ExecutionPattern.REPEAT.value:
            execution = cls.Execution.from_dict(
                {
                    'execution_pattern': execution_pattern,
                    'start_time': start_time,
                    'stop_time': stop_time,
                    'repeat_interval': execution_repeat_interval,
                }
            )

        recurrence_pattern = get_enum_val(
            recurrence_pattern, ScheduleEnums.RecurrencePattern
        )

        if recurrence_pattern == ScheduleEnums.RecurrencePattern.DAILY.value:
            daily = cls.Daily.from_dict(
                {'daily_pattern': daily_pattern, 'repeat_interval': repeat_interval}
            )
            return cls(
                recurrence_pattern=recurrence_pattern, execution=execution, daily=daily
            )

        if recurrence_pattern == ScheduleEnums.RecurrencePattern.WEEKLY.value:
            if days_of_week and not isinstance(days_of_week, list):
                helper.exception_handler(
                    msg='Error: days_of_week must be provided as a list',
                    exception_type=ValueError,
                )
            weekly = cls.Weekly.from_dict(
                {'repeat_interval': repeat_interval, 'days_of_week': days_of_week}
            )
            return cls(
                recurrence_pattern=recurrence_pattern,
                execution=execution,
                weekly=weekly,
            )

        if recurrence_pattern == ScheduleEnums.RecurrencePattern.MONTHLY.value:
            monthly = cls.Monthly.from_dict(
                {
                    "monthly_pattern": monthly_pattern,
                    "repeatInterval": repeat_interval,
                    "day": day,
                    "week_offset": week_offset,
                    "day_of_week": day_of_week,
                    "WeekdayOffset": weekday_offset,
                    "days_of_month": days_of_month,
                }
            )
            return cls(
                recurrence_pattern=recurrence_pattern,
                execution=execution,
                monthly=monthly,
            )

        if recurrence_pattern == ScheduleEnums.RecurrencePattern.YEARLY.value:
            yearly = cls.Yearly.from_dict(
                {
                    'yearly_pattern': yearly_pattern,
                    'month': month,
                    'day': day,
                    'week_offset': week_offset,
                    'day_of_week': day_of_week,
                }
            )
            return cls(
                recurrence_pattern=recurrence_pattern,
                execution=execution,
                yearly=yearly,
            )

    def update_properties(
        self,
        recurrence_pattern: ScheduleEnums.RecurrencePattern | str | None = None,
        execution_pattern: ScheduleEnums.ExecutionPattern | str | None = None,
        execution_time: str | None = None,
        start_time: str | None = None,
        stop_time: str | None = None,
        execution_repeat_interval: int | None = None,
        daily_pattern: ScheduleEnums.DailyPattern | str | None = None,
        repeat_interval: int | None = None,
        days_of_week: list[ScheduleEnums.DaysOfWeek] | list[str] | None = None,
        day: int | None = None,
        month: int | None = None,
        week_offset: ScheduleEnums.WeekOffset | str | None = None,
        day_of_week: ScheduleEnums.DaysOfWeek | str | None = None,
        weekday_offset: ScheduleEnums.WeekdayOffset | str | None = None,
        days_of_month: list[str] | None = None,
        monthly_pattern: ScheduleEnums.MonthlyPattern | str | None = None,
        yearly_pattern: ScheduleEnums.YearlyPattern | str | None = None,
    ):
        """
        Updates ScheduleTime object according to provided parameters. If a
        provided parameter is not None then update it, else use its current
        value.

        Args:
            recurrence_pattern (ScheduleEnums.RecurrencePattern, optional):
                The recurrence pattern of the schedule. Possible values are
                DAILY, WEEKLY, MONTHLY, YEARLY. Defaults to None.
            execution_pattern (ScheduleEnums.ExecutionPattern, optional):
                The execution pattern of the schedule. Possible values are ONCE,
                REPEAT. Defaults to None.
            execution_time (str, optional):
                The execution time of the execution day, if execution_pattern is
                ONCE. Format should be HH:mm:ss. Defaults to ''.
            start_time (str, optional):
                The start time of the execution day, if execution_pattern is
                REPEAT. Format should be HH:mm:ss. Defaults to ''.
            stop_time (str, optional):
                The stop time of the execution day, if execution_pattern is
                REPEAT. Format should be HH:mm:ss. Defaults to ''.
            execution_repeat_interval (int, optional):
                The repeat interval of minutes of the execution day, if
                execution_pattern is REPEAT. Defaults to None.
            daily_pattern (ScheduleEnums.DailyPattern, optional):
                The daily recurrence pattern of the schedule. Possible values
                are DAY, WEEKDAY. Defaults to None.
            repeat_interval (int, optional):
                The repeat interval of days of daily schedule, if daily_pattern
                is DAY. Defaults to None.
            day (int, optional):
                The day in month of monthly schedule, if monthly_pattern is DAY
                or The day in month of yearly schedule, if yearly_pattern is
                DAY. Defaults to None.
            month (int, optional):
                The month in year of yearly schedule. Defaults to None.
            week_offset (ScheduleEnums.WeekOffset, optional):
                The week offset in month of monthly schedule, if monthly_pattern
                is DAY_OF_WEEK or, The week offset in year of yearly schedule,
                if yearly_pattern is DAY_OF_WEEK. Possible values are FIRST,
                SECOND, THIRD, FOURTH, LAST. Defaults to None.
            day_of_week (ScheduleEnums.DaysOfWeek, optional):
                The day of week in month of monthly schedule, if monthly_pattern
                is DAY_OF_WEEK or, The day of week in year of yearly schedule,
                if yearly_pattern is DAY_OF_WEEK. Possible values are: MONDAY,
                TUESDAY, WEDNESDAY, THURSDAY, FRIDAY, SATURDAY, SUNDAY.
                Defaults to None.
            days_of_week (List[ScheduleEnums.DaysOfWeek], optional):
                The days of week of weekly schedule. Possible values are:
                MONDAY, TUESDAY, WEDNESDAY, THURSDAY, FRIDAY, SATURDAY, SUNDAY.
                Defaults to None.
            weekday_offset (ScheduleEnums.WeekdayOffset, optional):
                The weekday offset in month of monthly schedule, if
                monthly_pattern is WEEKDAY. Defaults to None.
            days_of_month (List[str], optional):
                The days of month of monthly schedule, if monthly_pattern is
                DAYS_OF_MONTH. Must be provided as a list of one or more
                stringified digits (from '1' to '31'). Defaults to None.
            monthly_pattern (ScheduleEnums.MonthlyPattern, optional):
                The monthly recurrence pattern of the schedule. Possible values
                are: DAY, DAY_OF_WEEK, WEEKDAY, LAST_DAY, DAYS_OF_MONTH.
                Defaults to None.
            yearly_pattern (ScheduleEnums.YearlyPattern, optional):
                The yearly recurrence pattern of the schedule. Possible values
                are DAY, DAY_OF_WEEK. Defaults to None.
        """

        if execution_pattern:
            execution_pattern = (
                execution_pattern
                if isinstance(execution_pattern, ScheduleEnums.ExecutionPattern)
                else ScheduleEnums.ExecutionPattern(execution_pattern)
            )
        self.execution.execution_pattern = (
            execution_pattern or self.execution.execution_pattern
        )
        if recurrence_pattern:
            recurrence_pattern = (
                recurrence_pattern
                if isinstance(recurrence_pattern, ScheduleEnums.RecurrencePattern)
                else ScheduleEnums.RecurrencePattern(recurrence_pattern)
            )
        self.recurrence_pattern = recurrence_pattern or self.recurrence_pattern

        if self.execution.execution_pattern == ScheduleEnums.ExecutionPattern.ONCE:
            self.execution = self.Execution.from_dict(
                {
                    'execution_pattern': execution_pattern
                    or self.execution.execution_pattern,
                    'execution_time': execution_time or self.execution.execution_time,
                }
            )
        elif self.execution.execution_pattern == ScheduleEnums.ExecutionPattern.REPEAT:
            self.execution = self.Execution.from_dict(
                {
                    'execution_pattern': execution_pattern
                    or self.execution.execution_pattern,
                    'start_time': start_time or self.execution.start_time,
                    'stop_time': stop_time or self.execution.stop_time,
                    'repeat_interval': execution_repeat_interval
                    or self.execution.repeat_interval,
                }
            )

        if self.recurrence_pattern == ScheduleEnums.RecurrencePattern.DAILY:
            if hasattr(self, 'daily') and self.daily:
                self.daily = self.Daily.from_dict(
                    {
                        'daily_pattern': daily_pattern or self.daily.daily_pattern,
                        'repeat_interval': (
                            repeat_interval
                            or getattr(self.daily, 'repeat_interval', None)
                        ),
                    }
                )
            else:
                self.daily = self.Daily.from_dict(
                    {'daily_pattern': daily_pattern, 'repeat_interval': repeat_interval}
                )

        if self.recurrence_pattern == ScheduleEnums.RecurrencePattern.WEEKLY:
            if hasattr(self, 'weekly') and self.weekly:
                self.weekly = self.Weekly.from_dict(
                    {
                        'repeat_interval': repeat_interval
                        or self.weekly.repeat_interval,
                        'days_of_week': days_of_week or self.weekly.days_of_week,
                    }
                )
            else:
                self.weekly = self.Weekly.from_dict(
                    {'repeat_interval': repeat_interval, 'days_of_week': days_of_week}
                )

        if self.recurrence_pattern == ScheduleEnums.RecurrencePattern.MONTHLY:
            if hasattr(self, 'monthly') and self.monthly:
                self.monthly = self.Monthly.from_dict(
                    {
                        "monthly_pattern": monthly_pattern
                        or self.monthly.monthly_pattern,
                        "repeat_interval": repeat_interval
                        or self.monthly.repeat_interval,
                        "day": day or getattr(self.monthly, 'day', None),
                        "week_offset": week_offset
                        or getattr(self.monthly, 'week_offset', None),
                        "day_of_week": day_of_week
                        or getattr(self.monthly, 'day_of_week', None),
                        "weekday_offset": (
                            weekday_offset
                            or getattr(self.monthly, 'weekday_offset', None)
                        ),
                        "days_of_month": days_of_month
                        or getattr(self.monthly, 'days_of_month', None),
                    }
                )
            else:
                self.monthly = self.Monthly.from_dict(
                    {
                        "monthly_pattern": monthly_pattern,
                        "repeat_interval": repeat_interval,
                        "day": day,
                        "week_offset": week_offset,
                        "day_of_week": day_of_week,
                        "weekday_offset": weekday_offset,
                        "days_of_month": days_of_month,
                    }
                )

        if self.recurrence_pattern == ScheduleEnums.RecurrencePattern.YEARLY:
            if hasattr(self, 'yearly') and self.yearly:
                self.yearly = self.Yearly.from_dict(
                    {
                        'yearly_pattern': yearly_pattern or self.yearly.yearly_pattern,
                        'month': month or getattr(self.yearly, 'month', None),
                        'day': day or getattr(self.yearly, 'day', None),
                        'week_offset': week_offset
                        or getattr(self.yearly, 'week_offset', None),
                        'day_of_week': day_of_week
                        or getattr(self.yearly, 'day_of_week', None),
                    }
                )
            else:
                self.yearly = self.Yearly.from_dict(
                    {
                        'yearly_pattern': yearly_pattern,
                        'month': month,
                        'day': day,
                        'week_offset': week_offset,
                        'day_of_week': day_of_week,
                    }
                )



"""
Create a dataframe from a Report.
List available reports, retrieve their attributes and metrics. Operate with
given report.

1. Connect to the environment using data from workstation
2. List available reports (with limit of 10)
3. Get report with given id and check its available attributes and metrics
4. Choose attributes and metrics for filtering
5. Apply filters to the report specifying attributes and metrics
6. Create a dataframe from the filtered report
7. Print first few rows of newly created dataframe
"""

from mstrio.connection import get_connection
from mstrio.project_objects import Report, Prompt, list_reports

# connect to environment without providing user credentials
# variable `workstationData` is stored within Workstation
connection = get_connection(workstationData, project_name='MicroStrategy Tutorial')

# List available Reports (limited to 10)
sample_reports = list_reports(connection, limit=10)
for report in sample_reports:
    print(report)

# Check available attributes and metrics of a Report
sample_report_id = '759C0B5340E0AD6FAEDFF19D4DBC3488'
sample_report = Report(connection, id=sample_report_id)
attributes = sample_report.attributes
metrics = sample_report.metrics
print(f"Attributes: {attributes}\nMetrics: {metrics}")

# Choose attributes and metrics
category_attribute = attributes[0].get('id')
subcategory_attribute = attributes[1].get('id')

cost_metric = metrics[0].get('id')
profit_metric = metrics[1].get('id')

# Filter which attributes and metrics to use in a dataframe
sample_report.apply_filters(
    attributes=[category_attribute, subcategory_attribute],
    metrics=[cost_metric, profit_metric],
)

# Create a dataframe from a Report
dataframe = sample_report.to_dataframe()
print(dataframe.head())

# Set and view VLDB properties for the report
vldb_properties = sample_report.list_vldb_settings()
sample_report.alter_vldb_settings(property_set_name='VLDB Select',name='Base Table Join for Template',value=2)
print(vldb_properties)

# View Owner/ACL properties for the report
acls = sample_report.list_acl()
owner_properties = sample_report.owner.list_properties().get('name')
print(f"ACLs: {acls}\nOwner_property 'name': '{owner_properties}'")


# Page by operations for report
page_by_report = Report(connection, '918EFA5F46E541B9F2799BA36E184811')

# List available elements
pg_elements = page_by_report.page_by_elements
print(pg_elements)


ATTR_QUARTER = '8D679D4A11D3E4981000E787EC6DE8A4'
ATTR_CC = '8D679D3511D3E4981000E787EC6DE8A4'
element_quarter = pg_elements[ATTR_QUARTER][0]

page_by_dataframe = page_by_report.to_dataframe(page_element_id={ATTR_QUARTER: element_quarter, ATTR_CC: 'h1;'})
print(page_by_dataframe.head())


# Prompt operations for report
prompt_report = Report(connection=connection, id='87FDA2E94FCD7C00B0A43AA2C52767B4')
prompt = Prompt(type='VALUE', key='CA6906D3499B6AEE259BFE9C308076D7@0@10', answers=100, use_default=False)

prompt_report_df = prompt_report.to_dataframe(prompt_answers=[prompt])
print("Prompt report dataframe:\n", prompt_report_df.head())



"""This is the demo script to show how administrator can manage projects
and its settings.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""
from mstrio.connection import get_connection
from mstrio.datasources import DatasourceInstance
from mstrio.server import DuplicationConfig, Environment, Project, compare_project_settings, list_languages
from mstrio.server.project import LockType
from mstrio.server.setting_types import FailedEmailDelivery

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name  # Insert name of project to interact with

conn = get_connection(workstationData, project_name=PROJECT_NAME)
env = Environment(connection=conn)

# get list of all projects or those just loaded
all_projects = env.list_projects()
loaded_projects = env.list_loaded_projects()

# get project with a given name
project = Project(connection=conn, name=PROJECT_NAME)

# get nodes from the project
nodes = project.nodes

# Define a variable which can be later used in a script
NODE_NAME = $node_name  # Insert name of node, which is in a project, to load or unload it

# idle/resume a project on a given node(s) (name(s) of existing node(s)
# should be used)
project.idle(on_nodes=[NODE_NAME])
project.resume(on_nodes=NODE_NAME)
# unload/load a project from/to a given node(s) (name(s) of existing
# node(s) should be used)
project.unload(on_nodes=NODE_NAME)
project.load(on_nodes=NODE_NAME)

# lock/unlock a project and get information about the lock
project.lock(lock_type=LockType.TEMPORAL_INDIVIDUAL)
# fetch lock status of a project
project.fetch('lock_status')
# get lock status of a project
lock_type = project.lock_status.lock_type
lock_time = project.lock_status.lock_time
lock_owner = project.lock_status.owner

project.unlock(lock_type=LockType.TEMPORAL_INDIVIDUAL, force=True)

# get settings of a project as a dataframe
project_settings_df = project.settings.to_dataframe()

# duplicate a project by default
DUPLICATION_NAME = $duplication_name  # Insert name of duplicated project
project.duplicate(target_name=DUPLICATION_NAME)

# duplicate the project with no default settings
# list languages to select from in duplication settings
langs = list_languages(conn)

selected_langs = langs[:5]  # Select first 5 languages for duplication
selected_default_project_lang = langs[0].lcid  # Select first language as default locale

# Define not default duplication settings
duplication_config= DuplicationConfig(schema_objects_only=False,
                                      skip_empty_profile_folders=False,
                                      include_user_subscriptions=False,
                                      include_contact_subscriptions=False,
                                      import_description="test duplication with languages",
                                      import_default_locale=selected_default_project_lang,
                                      import_locales=[lang.lcid for lang in selected_langs])

# perform duplication
project.duplicate(target_name=(DUPLICATION_NAME + '_non_default'),
                  duplication_config= duplication_config)

# Define variables which can be later used in a script
PROJECT_DESCRIPTION = $project_description  # Insert description of newly created project
PROJECT_NEW_DESCRIPTION = $project_new_description

# create a project and store it in variable to have immediate access to it
new_project = env.create_project(name=PROJECT_NAME, description=PROJECT_DESCRIPTION)

# change description of a newly created project
new_project.alter(description=PROJECT_DESCRIPTION)

# delete a project

# first, the project need to be unloaded
new_project.unload()
# then delete the unloaded project,
# confirmation prompt will appear asking for a project name
new_project.delete()

# if a project is already unloaded, to delete it, it has to be retrieved
# via Environment object

# create an environment object
env = Environment(conn)
# list projects in the environment
projects = env.list_projects()
# get a project that you want to delete
project = projects['<index of a project to delete>']
# delete a project, confirmation prompt will appear asking for a project name
project.delete()


# Define variables which can be later used in a script
PROJECT_1_NAME = $project_1_name
PROJECT_2_NAME = $project_2_name

# compare settings of 2 projects (only differences will be stored into
# a DataFrame)
project1 = Project(connection=conn, name=PROJECT_1_NAME)
project2 = Project(connection=conn, name=PROJECT_2_NAME)
df_cmp = compare_project_settings(projects=[project, project1, project2], show_diff_only=True)

# Get the list of all settings and their values
project_settings = project.settings.list_properties()
# Get the list of all settings and their values with the settings description
project_settings_with_description = project.settings.list_properties(show_description=True)
# Get the list of all settings configuration (type, min/max value, options, etc.)
project_settings_types = project.settings.setting_types

# Define variables which can be later used in a script
CSV_FILE_EXPORT = $csv_file_export  # Name of file to which settings are exported
CSV_FILE_IMPORT = $csv_file_import  # Name of file from which settings are imported

# save/load settings of a project to/from a file (format can be 'csv',
# 'json' or 'pickle')
project.settings.to_csv(name=CSV_FILE_EXPORT)
# Export settings of a project to a csv file with the settings description
project.settings.to_csv(name=CSV_FILE_EXPORT, show_description=True)
project.settings.import_from(file=CSV_FILE_IMPORT)

# change a setting of a project
project.settings.cubeIndexGrowthUpperBound = 501

# Change setting with the use of Enum
project.settings.appendInfoForEmailDelivery = [
    FailedEmailDelivery.SUBSCRIPTION_NAME,
    FailedEmailDelivery.SUBSCRIPTION_RECIPIENT_NAME,
    FailedEmailDelivery.DELIVERY_STATUS,
    FailedEmailDelivery.DELIVERY_EMAIL_ADDRESS,
]
# Update multiple settings at once
project.settings.alter(maxEmailSubscriptionCount=100, maxExecutingJobsPerUser=50)
project.settings.update()

# Get the list of all cache settings and their values
project_cache_settings = project.settings.list_caching_properties()
# Get the list of all cache settings and their values with the settings description
project_cache_settings_with_description = project.settings.list_caching_properties(show_description=True)

# change a cache setting of a project
project.settings.maxReportCacheMemoryConsumption = 250

# update multiple cache settings at once
project.settings.alter(maxReportCacheMemoryConsumption=250, maxReportCacheCount=5000)
project.settings.update()

# delete object and element cache of a project
project.settings.delete_object_cache()
project.settings.delete_element_cache()

# delete object and element cache for all projects on environment
env.delete_server_object_cache()
env.delete_server_element_cache()

# print currently set value of Unified Quoting Behavior VldbSetting
print(project.vldb_settings['Quoting Behavior'].value)
# the same setting can be accessed using setting name from WS (display_name)
print(project.vldb_settings['Unified Quoting Behavior'].value)
# enable Unified Quoting Behavior using name as key
project.alter_vldb_settings(names_to_values={'Quoting Behavior': 1})
# disable Unified Quoting Behavior using display_name as key
project.alter_vldb_settings(names_to_values={'Unified Quoting Behavior': 0})
# edit Cartesian Join Governing setting
project.alter_vldb_settings(names_to_values={'Cartesian Join Governing': 1})

# print currently available data engine versions
print(project.get_data_engine_versions())
# print data engine versions connected to environment projects
for project in env.list_projects():
    print(project.vldb_settings['AEVersion'].value)
# update data engine versions for all projects on environment
NEW_DATA_ENGINE_VERSION = $new_data_engine_version # Insert number related to new data engine version
for project in env.list_projects():
    project.update_data_engine_version(NEW_DATA_ENGINE_VERSION)

# Define variables which can be later used in a script
DATASOURCE_ID = $datasource_id  # Insert ID of datasource to add to project
# Get datasource instance with a given ID
datasource = DatasourceInstance(connection=conn, id=DATASOURCE_ID)
# Add datasource to project
project.add_datasource(datasources=[datasource])
# Remove datasource from project
project.remove_datasource(datasources=[datasource])



"""List user privileges for all active users.

1. Connect to the environment using data from workstation
2. Get all active users
3. For each of retrieved users prepare dictionary with user's id, name,
   username and list of privileges and save it to table
4. For each of retrieved users print user's name, username and each of its
   privileges' name
"""

from typing import List

from mstrio.connection import Connection, get_connection
from mstrio.users_and_groups import list_users


def list_active_user_privileges(connection: "Connection") -> List[dict]:
    """List user privileges for all active users.

    Args:
        connection: Strategy One connection object returned by
            `connection.Connection()`

    Returns:
        list of dicts where each of them is in given form:
        {
            'id' - id of user
            'name' - name of user
            'username' - username of user
            'privileges' - list of privileges of user
        }
    """
    all_users = list_users(connection=connection)
    active_users = [u for u in all_users if u.enabled]
    privileges_list = []
    for usr in active_users:
        p = {
            'id': usr.id,
            'name': usr.name,
            'username': usr.username,
            'privileges': usr.privileges,
        }
        print(f"{p['name']} ({p['username']}) ", flush=True)
        for prvlg in p['privileges']:
            print("\t" + prvlg['privilege']['name'], flush=True)
        privileges_list.append(p)
    return privileges_list


# connect to environment without providing user credentials
# variable `workstationData` is stored within Workstation
conn = get_connection(workstationData, 'MicroStrategy Tutorial')

# list privileges for all enabled users
list_active_user_privileges(conn)



from typing import TYPE_CHECKING

from mstrio.modeling.expression import ExpressionFormat
from mstrio.utils.error_handlers import ErrorHandler
from mstrio.utils.helper import get_enum_val
from mstrio.utils.version_helper import is_server_min_version

if TYPE_CHECKING:
    from mstrio.connection import Connection
    from mstrio.utils.sessions import FuturesSessionWithRenewal

CUBE_FIELDS = '-data.metricValues.extras,-data.metricValues.formatted'


@ErrorHandler(err_msg="Error getting cube {id} definition.")
def cube_definition(connection: 'Connection', id: str):
    """Get the definition of a specific cube, including attributes and metrics.
    The cube can be either an Intelligent Cube or a Direct Data Access
    (DDA)/MDX cube. The in-memory cube definition provides information about
    all available objects without actually running any data query/report. The
    results can be used by other requests to help filter large datasets and
    retrieve values dynamically, helping with performance and scalability.

    Args:
        connection: Strategy One REST API connection object.
        id (str): Unique ID of the cube you wish to extract information
            from.

    Returns:
        Complete HTTP response object.
    """
    connection._validate_project_selected()
    return connection.get(endpoint=f'/api/v2/cubes/{id}')


@ErrorHandler(err_msg="Error getting cube {id} metadata information.")
def cube_info(connection: 'Connection', id: str):
    """Get information for specific cubes in a specific project. The cubes can
    be either Intelligent cubes or Direct Data Access (DDA)/MDX cubes. This
    request returns the cube name, ID, size, status, path, last modification
    date, and owner name and ID.

    Args:
        connection: Strategy One REST API connection object.
        id (str): Unique ID of the cube you wish to extract information
            from.

    Returns:
        Complete HTTP response object.
    """
    return connection.get(endpoint=f'/api/cubes/?id={id}')


@ErrorHandler(err_msg="Error getting cube {id} metadata information.")
def get_cube_status(connection: 'Connection', id: str):
    """Get the status of a specific cube in a specific project.

    Args:
        connection: Strategy One REST API connection object.
        id (str): Unique ID of the cube you wish to extract information
            from.

    Returns:
        Complete HTTP response object.
    """
    return connection.head(endpoint=f'/api/cubes/{id}')


@ErrorHandler(err_msg="Error creating a new cube instance with ID {cube_id}.")
def cube_instance(
    connection: 'Connection',
    cube_id: str,
    body: dict = None,
    offset: int = 0,
    limit: int = 5000,
):
    """Create a new instance of a specific cube. This in-memory instance can be
    used by other requests.

    Args:
        connection: Strategy One REST API connection object.
        cube_id (str): Unique ID of the cube you wish to extract information
            from.
        offset (int, optional): Starting point within the collection of returned
            results. Default is 0.
        limit (int, optional): Used to control data extract behavior on datasets
            which have a large number of rows. The default is 1000. As an
            example, if the dataset has 50,000 rows, this function will
            incrementally extract all 50,000 rows in 1,000 row chunks. Depending
            on system resources, using a higher limit setting (e.g. 10,000) may
            reduce the total time required to extract the entire dataset.

    Returns:
        Complete HTTP response object.
    """
    body = body or {}
    params = {'offset': offset, 'limit': limit}
    if is_server_min_version(connection, '11.2.0200'):
        params['fields'] = CUBE_FIELDS

    return connection.post(
        endpoint=f'/api/v2/cubes/{cube_id}/instances', json=body, params=params
    )


@ErrorHandler(err_msg="Error getting cube {cube_id} contents.")
def cube_instance_id(
    connection: 'Connection',
    cube_id: str,
    instance_id: str,
    offset: int = 0,
    limit: int = 5000,
):
    """Get the results of a previously created instance for a specific cube,
    using the in-memory instance created by cube_instance().

    Args:
        connection: Strategy One REST API connection object.
        cube_id (str): Unique ID of the cube you wish to extract information
            from.
        instance_id (str): Unique ID of the in-memory instance of a published
            cube.
        offset (int, optional): Starting point within the collection of returned
            results. Default is 0.
        limit (int, optional): Used to control data extract behavior on datasets
            which have a large number of rows. The default is 1000. As an
            example, if the dataset has 50,000 rows, this function will
            incrementally extract all 50,000 rows in 1,000 row chunks. Depending
            on system resources, using a higher limit setting (e.g. 10,000) may
            reduce the total time required to extract the entire dataset.

    Returns:
        Complete HTTP response object.
    """
    params = {'offset': offset, 'limit': limit}
    if is_server_min_version(connection, '11.2.0200'):
        params['fields'] = CUBE_FIELDS

    return connection.get(
        endpoint=f'/api/v2/cubes/{cube_id}/instances/{instance_id}', params=params
    )


def cube_instance_id_coroutine(
    future_session: 'FuturesSessionWithRenewal',
    cube_id: str,
    instance_id: str,
    offset: int = 0,
    limit: int = 5000,
):
    """Get the future of a previously created instance for a specific cube
    asynchronously, using the in-memory instance created by cube_instance().

    Returns:
        Complete Future object.
    """
    params = {'offset': offset, 'limit': limit}
    if is_server_min_version(future_session.connection, '11.2.0200'):
        params['fields'] = CUBE_FIELDS

    endpoint = f'/api/v2/cubes/{cube_id}/instances/{instance_id}'
    return future_session.get(endpoint=endpoint, params=params)


@ErrorHandler(
    err_msg="Error getting attribute {attribute_id} elements within cube {cube_id}"
)
def cube_single_attribute_elements(
    connection: 'Connection',
    cube_id: str,
    attribute_id: str,
    offset: int = 0,
    limit: int = 200000,
):
    """Get elements of a specific attribute of a specific cube.

    Args:
        connection: Strategy One REST API connection object.
        cube_id (str): Unique ID of the cube you wish to extract information
            from.
        attribute_id (str): Unique ID of the attribute in the cube.

    Returns:
        Complete HTTP response object.
    """

    return connection.get(
        endpoint=f'/api/cubes/{cube_id}/attributes/{attribute_id}/elements',
        params={'offset': offset, 'limit': limit},
    )


def cube_single_attribute_elements_coroutine(
    future_session: 'FuturesSessionWithRenewal',
    cube_id: str,
    attribute_id: str,
    offset: int = 0,
    limit: int = 200000,
):
    """Get elements of a specific attribute of a specific cube.

    Returns:
        Complete Future object.
    """
    endpoint = f'/api/cubes/{cube_id}/attributes/{attribute_id}/elements'
    return future_session.get(
        endpoint=endpoint, params={'offset': offset, 'limit': limit}
    )


@ErrorHandler(err_msg="Error sending request to publish cube with ID {cube_id}")
def publish(connection: 'Connection', cube_id: str):
    """Publish a specific cube in a specific project.

    Args:
        connection: Strategy One REST API connection object.
        cube_id (str): Unique ID of the cube you wish to publish.

    Returns:
        Complete HTTP response object.
    """

    return connection.post(endpoint=f'/api/v2/cubes/{cube_id}')


@ErrorHandler(err_msg="Error getting cube {cube_id} status.")
def status(connection: 'Connection', cube_id: str, throw_error: bool = True):
    """Get the status of a specific cube in a specific project. The status is
    returned in HEADER X-MSTR-CubeStatus with a value from EnumDSSCubeStates,
    which is a bit vector.

    Args:
        connection: Strategy One REST API connection object.
        cube_id (str): Unique ID of the cube you wish to extract information
            from.
        throw_error (bool, optional): Flag indicates if the error should be
            thrown.

    Returns:
        Complete HTTP response object.
    """

    return connection.head(endpoint=f'/api/cubes/{cube_id}')


@ErrorHandler(err_msg="Error creating cube {name} definition.")
def create(
    connection: 'Connection',
    name: str,
    folder_id: str,
    overwrite: bool = None,
    description: str = None,
    definition: dict = None,
):
    """
    Create an intelligent cube.
    POST /api/v2/cubes
    """
    connection._validate_project_selected()

    body = {
        'name': name,
        'description': description,
        'folderId': folder_id,
        'overwrite': overwrite,
        'definition': definition,
    }
    params = {'X-MSTR-ProjectID': connection.project_id}

    return connection.post(endpoint='/api/v2/cubes', json=body, params=params)


@ErrorHandler(err_msg="Error updating cube {cube_id} definition.")
def update(connection: 'Connection', cube_id: str, definition: dict = None):
    """
    Update an intelligent cube.
    PUT /api/v2/cubes/{cube_id}
    """
    connection._validate_project_selected()

    body = {'definition': definition}
    params = {'X-MSTR-ProjectID': connection.project_id}

    return connection.put(endpoint=f'/api/v2/cubes/{cube_id}', json=body, params=params)


@ErrorHandler(err_msg="Error getting sql view of cube with ID {cube_id}")
def get_sql_view(connection: 'Connection', cube_id: str, project_id: str = None):
    """
    Get the sql view of cube.
    GET /api/v2/cubes/{cube_id}/sqlView
    """
    if not project_id:
        connection._validate_project_selected()
        project_id = connection.project_id

    return connection.get(
        endpoint=f'/api/v2/cubes/{cube_id}/sqlView',
        params={'X-MSTR-projectID': project_id},
    )


@ErrorHandler(err_msg="Error creating cube {name}.")
def create_cube(
    connection: 'Connection',
    body: dict,
    project_id: str | None = None,
    cube_template_id: str | None = None,
    show_expression_as: ExpressionFormat | str = ExpressionFormat.TREE,
    show_filter_tokens: bool = False,
    show_advanced_properties: bool = True,
):
    """Create a new cube.

    Args:
        connection (Connection): Strategy One REST API connection object.
        body (dict): JSON-formatted data used to create cube.
        project_id (str, optional): Project ID.
        cube_template_id (str, optional): If specified, new created cube will
            inherit 'template' of given by ID cube.
        show_expression_as (ExpressionFormat, str, optional): Specify how
            expressions should be presented, default 'ExpressionFormat.TREE'.
            Available values:
            - None.
                (expression is returned in 'text' format)
            - `ExpressionFormat.TREE` or `tree`.
                (expression is returned in `text` and `tree` formats)
            - `ExpressionFormat.TOKENS` or `tokens`.
                (expression is returned in `text` and `tokens` formats)
        show_filter_tokens (bool, optional): Specify whether 'filter'
            is returned in 'tokens' format, along with `text` and `tree`
            formats, default False.
            - If omitted or False, only `text` and `tree` formats are returned.
            - If True, all `text`, 'tree' and `tokens` formats are returned.
        show_advanced_properties (bool, optional): Specify whether to retrieve
            the values of the advanced properties. If omitted or false, nothing
            will be returned for the advanced properties, default True.

    Returns:
        HTTP response object. Expected status: 200.

    Raises:
        AttributeError: If project ID is not provided and Connection object
            doesn't have project selected.
    """

    if project_id is None:
        connection._validate_project_selected()
        project_id = connection.project_id

    params = {
        'cubeTemplateId': cube_template_id,
        'showExpressionAs': get_enum_val(show_expression_as, ExpressionFormat),
        'showFilterTokens': str(show_filter_tokens).lower(),
        'showAdvancedProperties': str(show_advanced_properties).lower(),
    }

    return connection.post(
        endpoint='/api/model/cubes',
        headers={'X-MSTR-ProjectID': project_id},
        json=body,
        params=params,
    )


@ErrorHandler(err_msg="Error getting cube with ID {id}")
def get_cube(
    connection: 'Connection',
    id: str,
    project_id: str | None = None,
    show_expression_as: ExpressionFormat | str = ExpressionFormat.TREE,
    show_filter_tokens: bool = False,
    show_advanced_properties: bool = True,
):
    """Get cube by specific ID.

    Args:
        connection (Connection): Strategy One REST API connection object.
        id (str): Cube ID.
        project_id (str, optional): Project ID.
        show_expression_as (ExpressionFormat, str, optional): Specify how
            expressions should be presented, default 'ExpressionFormat.TREE'.
            Available values:
            - None.
                (expression is returned in 'text' format)
            - `ExpressionFormat.TREE` or `tree`.
                (expression is returned in `text` and `tree` formats)
            - `ExpressionFormat.TOKENS` or `tokens`.
                (expression is returned in `text` and `tokens` formats)
        show_filter_tokens (bool, optional): Specify whether 'filter'
            is returned in 'tokens' format, along with `text` and `tree`
            formats, default False.
            - If omitted or False, only `text` and `tree` formats are returned.
            - If True, all `text`, 'tree' and `tokens` formats are returned.
        show_advanced_properties (bool, optional): Specify whether to retrieve
            the values of the advanced properties. If omitted or false, nothing
            will be returned for the advanced properties, default True.

    Returns:
        HTTP response object. Expected status: 200.

    Raises:
        AttributeError: If project ID is not provided and Connection object
            doesn't have project selected.
    """

    if project_id is None:
        connection._validate_project_selected()
        project_id = connection.project_id

    params = {
        'showExpressionAs': get_enum_val(show_expression_as, ExpressionFormat),
        'showFilterTokens': str(show_filter_tokens).lower(),
        'showAdvancedProperties': str(show_advanced_properties).lower(),
    }

    return connection.get(
        endpoint=f'/api/model/cubes/{id}',
        headers={'X-MSTR-ProjectID': project_id},
        params=params,
    )


@ErrorHandler(err_msg="Error updating cube with ID {id}")
def update_cube(
    connection: 'Connection',
    id: str,
    body: dict,
    project_id: str | None = None,
    show_expression_as: ExpressionFormat | str = ExpressionFormat.TREE,
    show_filter_tokens: bool = False,
    show_advanced_properties: bool = False,
):
    """Update cube specified by ID.

    Args:
        connection (Connection): Strategy One REST API connection object.
        id (str): Cube ID.
        body (dict): JSON-formatted data used to update cube.
        project_id (str, optional): Project ID.
        show_expression_as (ExpressionFormat, str, optional): Specify how
            expressions should be presented, default 'ExpressionFormat.TREE'.
            Available values:
            - None.
                (expression is returned in 'text' format)
            - `ExpressionFormat.TREE` or `tree`.
                (expression is returned in `text` and `tree` formats)
            - `ExpressionFormat.TOKENS` or `tokens`.
                (expression is returned in `text` and `tokens` formats)
        show_filter_tokens (bool, optional): Specify whether 'filter'
            is returned in 'tokens' format, along with `text` and `tree`
            formats, default False.
            - If omitted or False, only `text` and `tree` formats are returned.
            - If True, all `text`, 'tree' and `tokens` formats are returned.
        show_advanced_properties (bool, optional): Specify whether to retrieve
            the values of the advanced properties. If omitted or false, nothing
            will be returned for the advanced properties, default False.

    Returns:
        HTTP response object. Expected status: 200.

    Raises:
        AttributeError: If project ID is not provided and Connection object
            doesn't have project selected.
    """

    if project_id is None:
        connection._validate_project_selected()
        project_id = connection.project_id

    params = {
        'showExpressionAs': get_enum_val(show_expression_as, ExpressionFormat),
        'showFilterTokens': str(show_filter_tokens).lower(),
        'showAdvancedProperties': str(show_advanced_properties).lower(),
    }

    return connection.put(
        endpoint=f'/api/model/cubes/{id}',
        headers={'X-MSTR-ProjectID': project_id},
        json=body,
        params=params,
    )


@ErrorHandler(err_msg="Error getting metadata of VLDB settings for cube with ID {id}")
def get_applicable_vldb_settings(
    connection: 'Connection', id: str, project_id: str | None = None
):
    """Get metadata of advanced VLDB settings for cube.

    Args:
        connection (Connection): Strategy One REST API connection object.
        id (str): Cube ID.
        project_id (str, optional): Project ID.

    Returns:
        HTTP response object. Expected status: 200.

    Raises:
        AttributeError: If project ID is not provided and Connection object
            doesn't have project selected.
    """

    if project_id is None:
        connection._validate_project_selected()
        project_id = connection.project_id

    return connection.get(
        endpoint=f'/api/model/cubes/{id}/applicableVldbProperties',
        headers={'X-MSTR-ProjectID': project_id},
    )



# values is a vector input
# Metric expression example: ScriptFunctionRelative<_InputNames="values",  _ScriptName="get_running_sums", BreakBy={Year}, 
#                            SortBy={Date}>(Revenue)

def get_running_sums(values):
    # Initialize an empty list to store the running sums
    running_sums = []
    # Initialize a variable to keep track of the cumulative sum
    cumulative_sum = 0
    # Iterate through the list of values
    for value in values:
        # Update the cumulative sum with the current value
        cumulative_sum += value
        # Append the cumulative sum to the running sums list
        running_sums.append(cumulative_sum)
    return running_sums

# must-have method which returns the values served as the metric evaluation results. 
def get_results():
    return get_running_sums($values)

print(get_results())



from typing import TYPE_CHECKING

from mstrio.utils.api_helpers import changeset_manager, unpack_information
from mstrio.utils.error_handlers import ErrorHandler

if TYPE_CHECKING:
    from mstrio.connection import Connection


@unpack_information
@ErrorHandler(err_msg="Error creating the filter.")
def create_filter(
    connection: 'Connection',
    body: dict,
    show_expression_as: str | None = None,
    show_filter_tokens: bool = False,
    error_msg: str | None = None,
):
    """Creates a new filter in the changeset,
    based on the definition provided in request body.

    Args:
        connection: Strategy One REST API connection object
        body (dict): Filter creation body
        error_msg (str, optional): Custom Error Message for Error Handling
        show_expression_as (str, optional): specify how expressions should be
            presented
            Available values:
                - `tree` (default)
                - `tokens`
        show_filter_tokens (bool, optional): specify whether `qualification`
            is returned in `tokens` format, along with `text` and `tree`
            format

    Returns:
        Complete HTTP response object. Expected status is 201.
    """
    with changeset_manager(connection) as changeset_id:
        return connection.post(
            endpoint='/api/model/filters',
            headers={'X-MSTR-MS-Changeset': changeset_id},
            json=body,
            params={
                'showExpressionAs': show_expression_as,
                'showFilterTokens': str(show_filter_tokens).lower(),
            },
        )


@unpack_information
@ErrorHandler(err_msg="Error getting the filter with ID: {id}.")
def get_filter(
    connection: 'Connection',
    id: str,
    project_id: str | None = None,
    changeset_id: str | None = None,
    show_expression_as: str | None = None,
    show_filter_tokens: bool = False,
    error_msg: str | None = None,
):
    """Get the definition of a filter.
    The project ID is required to return a filter's definition
    in metadata. The changeset ID is required to return a filter's
    definition within a specific changeset. To execute the request,
    either the project ID or changeset ID needs to be provided.
    If both are provided, only the changeset ID is used.

    Args:
        connection: Strategy One REST API connection object
        id (str): Filter ID. The ID can be:
            - the object ID used in the metadata.
            - the object ID used in the changeset, but not yet committed
            to metadata.
        project_id (str, optional): Project ID
        changeset_id (str, optional): Changeset ID
        error_msg (str, optional): Custom Error Message for Error Handling
        show_expression_as (str, optional): specify how expressions should be
            presented
            Available values:
                - `tree` (default)
                - `tokens`
        show_filter_tokens (bool, optional): specify whether `qualification`
            is returned in `tokens` format, along with `text` and `tree`
            format

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    if project_id is None:
        connection._validate_project_selected()
        project_id = connection.project_id

    return connection.get(
        endpoint=f'/api/model/filters/{id}',
        headers={'X-MSTR-ProjectID': project_id, 'X-MSTR-MS-Changeset': changeset_id},
        params={
            'showExpressionAs': show_expression_as,
            'showFilterTokens': str(show_filter_tokens).lower(),
        },
    )


@unpack_information
@ErrorHandler(err_msg="Error updating the filter with ID: {id}.")
def update_filter(
    connection: 'Connection',
    id: str,
    body: dict,
    show_expression_as: str | None = None,
    show_filter_tokens: bool = False,
    error_msg: str | None = None,
):
    """Updates a specific filter in the changeset,
    based on the definition provided in the request body.

    Args:
        connection: Strategy One REST API connection object
        id (str): Filter ID. The ID can be:
            - the object ID used in the metadata.
            - the object ID used in the changeset, but not yet committed
            to metadata.
        body (dict): Filter update info
        error_msg (str, optional): Custom Error Message for Error Handling
        show_expression_as (str, optional): specify how expressions should be
            presented
            Available values:
                - `tree` (default)
                - `tokens`
        show_filter_tokens (bool, optional): specify whether `qualification`
            is returned in `tokens` format, along with `text` and `tree`
            format

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    with changeset_manager(connection) as changeset_id:
        return connection.put(
            endpoint=f'/api/model/filters/{id}',
            headers={'X-MSTR-MS-Changeset': changeset_id},
            json=body,
            params={
                'showExpressionAs': show_expression_as,
                'showFilterTokens': str(show_filter_tokens).lower(),
            },
        )



# Create an object package using results of search for last modified objects
# Add dataset components for Dashboard object
# Set up action "Replace" to all of the objects

from datetime import UTC, datetime, timedelta

from mstrio.connection import get_connection


from mstrio.object_management.migration.migration import Migration
from mstrio.object_management.migration.package import (
    Action,
    PackageConfig,
    PackageContentInfo,
    PackageSettings,
)
from mstrio.object_management.search_operations import full_search
from mstrio.types import ObjectTypes

conn = get_connection(workstationData, project_name='Project Analytics')

DATE_CUTOFF = datetime.now(tz=UTC) - timedelta(days=30)
# format to str; example output: 2023-04-04T06:33:32Z
DATE_CUTOFF_STR = DATE_CUTOFF.strftime('%Y-%m-%dT%H:%M:%SZ')

search_results = full_search(
    conn, conn.project_id, begin_modification_time=DATE_CUTOFF_STR
)
datasets = []
for obj in search_results:
    if obj.type == ObjectTypes.DOCUMENT_DEFINITION:
        dependent_datasets = obj.list_dependents(object_types=[ObjectTypes.DBROLE])
        datasets += dependent_datasets

migrated_objects = search_results + datasets

# You can set the default action for all migrated objects
# using the PackageSettings object
package_settings = PackageSettings(Action.REPLACE)

# For each of the migrated objects, create a PackageContentInfo entry
# An entry can have its own action defined. It will override the default action
# If not specified, the default action will be used
# In this case, we will not set include_dependents to True. This will cause
# I-Server to include all dependent objects, without options to include or
# exclude them.
content_info = [
    PackageContentInfo(id=obj.id, type=obj.type, action=Action.REPLACE)
    for obj in migrated_objects
]

package_config = PackageConfig(package_settings, content_info)

my_obj_mig_alt = Migration.create_object_migration(
    connection=conn,
    toc_view=package_config,
    name="test_object_mig",
    project_id=conn.project_id,
)



"""List security roles for every user in a user group. It is possible to provide
either name or id of user group. Without any changes this script will be
executed for user group 'System Administrators'.

1. Connect to the environment using data from workstation
2. Get user group object based on provided name or id (in this case name of user
   group is 'System Administrators')
3. Get list of all members of user group
4. For each member check whether it is user or user group (if additional argument
   is set to True, then skip user groups)
5. Prepare dictionary with type, id, name, username and list of security roles
6. For each project print its name and name of every security role which is
   inside the given project for the given member of "main" user group
"""

from typing import List

from mstrio.connection import Connection, get_connection
from mstrio.users_and_groups import UserGroup


def list_security_roles_per_user(
    connection: "Connection",
    user_group_name: str = None,
    user_group_id: str = None,
    include_user_groups: bool = False,
) -> List[dict]:
    """List security roles for every user in a user group.
    It is possible to provide either name or id of user group.

    Args:
        connection: Strategy One connection object returned by
            `connection.Connection()`
        user_group_name (str): name of the user group
        user_group_id (str): id of the user group
        include_user_groups (bool): if True also user groups, which are inside
            provided user group, will be included in the result

    Returns:
        list of dicts where each of them is in given form:
        {
            'type' - type of object (it can be `user` or `user group`)
            'id' - id of object
            'name' - name of object
            'username' - username of user (for user group it is None)
            'security_roles' - list of security roles which object has
        }
    """

    user_group_ = UserGroup(
        connection=connection, name=user_group_name, id=user_group_id
    )
    all_security_roles = []
    for member in user_group_.list_members():
        if isinstance(member, UserGroup):
            if not include_user_groups:
                continue
            member_type = 'user_group'
        else:
            member_type = 'user'
        m = {
            'type': member_type,
            'id': member.id,
            'name': member.name,
            'username': member.to_dict().get('username', None),
            'security_roles': member.security_roles,
        }
        print('Security roles:', flush=True)
        for project_ in m['security_roles']:
            print("\tProject: " + project_['name'], flush=True)
            for sec_role in project_['securityRoles']:
                print("\t\t" + sec_role['name'], flush=True)
        all_security_roles.append(m)
    return all_security_roles


# connect to environment without providing user credentials
# variable `workstationData` is stored within Workstation
conn = get_connection(workstationData, 'MicroStrategy Tutorial')

# list security roles per all users in user group 'System Administrators'
# change name of user group if needed
list_security_roles_per_user(conn, user_group_name='System Administrators')



import mstrio.api.calendars as calendars_api
from mstrio.connection import Connection
from mstrio.utils import helper


def _wrangle_year(source: dict, *keys) -> int | None:
    """Safely extract year from a nested dictionary, i.e.
    source[key1][key2]...[keyN]"""
    acc = source
    for key in keys:
        acc = acc.get(key, {})
    return int(acc) if acc else None


def _wrangle_entry(source: dict) -> dict:
    data = source.copy()
    if base_cal := data.get('baseCalendar'):
        data['baseCalendar'] = base_cal['objectId']

    data['calendarBeginStatic'] = _wrangle_year(data, 'calendarBegin', 'staticYear')
    data['calendarBeginOffset'] = _wrangle_year(
        data, 'calendarBegin', 'dynamicYearOffset'
    )
    data['calendarEndStatic'] = _wrangle_year(data, 'calendarEnd', 'staticYear')
    data['calendarEndOffset'] = _wrangle_year(data, 'calendarEnd', 'dynamicYearOffset')
    del data['calendarBegin']
    del data['calendarEnd']

    data['weekStartDay'] = data['weekStartDay'].lower()
    data = helper.camel_to_snake(data)

    return data


def get_calendar(connection: Connection, id: str) -> dict:
    data = calendars_api.get_calendar(connection, id).json()
    return _wrangle_entry(data)


def list_calendars(
    connection: Connection,
    subtype: str | None = None,
    limit: int | None = None,
    offset: int | None = None,
) -> list[dict]:
    data = calendars_api.list_calendars(
        connection=connection,
        subtype=subtype,
        offset=offset,
        limit=limit,
    ).json()['calendars']
    return [_wrangle_entry(entry) for entry in data]


def create_calendar(
    connection: Connection,
    body: dict,
):
    data = calendars_api.create_calendar(connection, body=body).json()
    return _wrangle_entry(data)


def update_calendar(
    connection: Connection,
    id: str,
    body: dict,
):
    data = calendars_api.update_calendar(connection, id, body=body).json()
    return _wrangle_entry(data)



from typing import TYPE_CHECKING

from mstrio.utils.error_handlers import ErrorHandler

if TYPE_CHECKING:
    from mstrio.connection import Connection
    from mstrio.utils.sessions import FuturesSessionWithRenewal


@ErrorHandler(err_msg="Error getting user groups privileges for a group with ID {id}")
def get_privileges(
    connection, id, privilege_level=None, project_id=None, error_msg=None
):
    """Get user group's privileges of a project including the source of the
    privileges.

    Args:
        connection: Strategy One REST API connection object
        id (string): ID of user group containing your required privileges
        privilege_level (string, optional) [server, project]: String
            representing level of privileges required
        project_id (string, optional): Project id string
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """

    return connection.get(
        endpoint=f'/api/usergroups/{id}/privileges',
        headers={'X-MSTR-ProjectID': None},
        params={'privilege.level': privilege_level, 'projectId': project_id},
    )


@ErrorHandler(err_msg="Error getting user groups memberships for a group with ID {id}")
def get_memberships(connection, id, error_msg=None):
    """Get information for the user group that is the direct parent of a
    specific user group.

    Args:
        connection: Strategy One REST API connection object
        id (string): ID of user group containing your required memberships
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """

    return connection.get(
        endpoint=f'/api/usergroups/{id}/memberships',
        headers={'X-MSTR-ProjectID': None},
    )


@ErrorHandler(err_msg="Error getting user groups members for a group with ID {id}")
def get_members(
    connection, id, include_access=False, offset=0, limit=-1, error_msg=None
):
    """Get member information for a specific user group.

    Args:
        connection: Strategy One REST API connection object
        id (string): ID of user group containing your required privileges
        include_access (bool, optional): Specifies whether to return access for
            members
        offset (integer, optional): Starting point within the collection of
            returned search results. Used to control paging behavior.
        limit (integer, optional): Maximum number of items returned for a single
            search request. Used to control paging behavior. Use -1 for no limit
            (subject to governing settings).
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """

    return connection.get(
        endpoint=f'/api/usergroups/{id}/members',
        headers={'X-MSTR-ProjectID': None},
        params={'includeAccess': include_access, 'offset': offset, 'limit': limit},
    )


@ErrorHandler(err_msg="Error getting user group settings for a group with ID {id}")
def get_settings(
    connection, id, include_access=False, offset=0, limit=-1, error_msg=None
):
    """Update the governing setting of the user group id.

    Args:
        connection: Strategy One REST API connection object
        id (string): ID of user group containing your required privileges
        include_access (bool, optional): Specifies whether to return access for
            members
        offset (integer, optional): Starting point within the collection of
            returned search results. Used to control paging behavior.
        limit (integer, optional): Maximum number of items returned for a single
            search request. Used to control paging behavior. Use -1 for no limit
            (subject to governing settings).
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        endpoint=f'/api/usergroups/{id}/settings',
        headers={'X-MSTR-ProjectID': None},
    )


@ErrorHandler(err_msg="Error getting user groups top level.")
def get_top_level(connection, error_msg=None):
    """Get information for all of the user groups that exist at the level of
    the Strategy One Everyone user group.

    Args:
        connection: Strategy One REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """

    return connection.get(
        endpoint='/api/usergroups/topLevel',
        headers={'X-MSTR-ProjectID': None},
    )


@ErrorHandler(err_msg="Error updating user group info with ID {id}")
def update_user_group_info(connection, id, body, error_msg=None):
    """Update specific information for a specific user group.

    Args:
        connection: Strategy One REST API connection object
        id (string): ID of user group containing your required privileges
        body (JSON): Body.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object
    """

    return connection.patch(
        endpoint=f'/api/usergroups/{id}',
        headers={'X-MSTR-ProjectID': None},
        json=body,
    )


@ErrorHandler(err_msg="Error deleting user group with ID {id}")
def delete_user_group(connection, id, error_msg=None):
    """Delete user group for specific user group id.

    Args:
        connection: Strategy One REST API connection object
        id (string): ID of user group containing your required privileges
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object
    """

    return connection.delete(
        endpoint=f'/api/usergroups/{id}',
        headers={'X-MSTR-ProjectID': None},
    )


@ErrorHandler(err_msg="Error overwriting user group info for a group with ID {id}")
def replace_user_group_info(connection, id, error_msg=None):
    """Update all of the information for a specific user group.

    Args:
        connection: Strategy One REST API connection object
        id (string): ID of user group containing your required privileges
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object
    """

    return connection.put(
        endpoint=f'/api/usergroups/{id}',
        headers={'X-MSTR-ProjectID': None},
    )


@ErrorHandler(err_msg="Error getting information for a set of user groups.")
def get_info_all_user_groups(
    connection, name_begins, offset=0, limit=1000, fields=None, error_msg=None
):
    """Get information for a specific set of user groups or all user groups.
    You can refine the set of user groups that are returned with a query
    parameter that specifies the characters that the user group name must begin
    with; if you omit the nameBegins query parameter, all user groups are
    returned.

    Args:
        connection (object): Strategy One connection object returned by
            `connection.Connection()`.
        name_begins (string): only user groups with given beginning of name will
            be retrieved
        offset (integer, optional): Starting point within the collection of
            returned search results. Used to control paging behavior.
        limit (integer, optional): Maximum number of items returned for a single
            search request. Used to control paging behavior. Use -1 for no limit
            (subject to governing settings).
        fields (list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object returned by the Strategy One REST server
    """

    return connection.get(
        endpoint='/api/usergroups/',
        headers={'X-MSTR-ProjectID': None},
        params={
            'nameBegins': name_begins,
            'offset': offset,
            'limit': limit,
            'fields': fields,
        },
    )


def get_info_all_user_groups_async(
    future_session: 'FuturesSessionWithRenewal',
    name_begins,
    offset=0,
    limit=-1,
    fields=None,
):
    """Get information for a set of users asynchronously.

    Args:
        future_session (FuturesSessionWithRenewal): Future Session object
            to call Strategy One REST Server asynchronously
        name_begins(string): Characters that the user name must begin with.
        offset(int): Starting point within the collection of returned search
            results. Used to control paging behavior.
        limit(int): Maximum number of items returned for a single search
            request. Used to control paging behavior. Use -1 (default ) for no
            limit (subject to governing settings).
        fields(list, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.

    Returns:
        HTTP response object returned by the Strategy One REST server.
    """

    params = {
        'nameBegins': name_begins,
        'offset': offset,
        'limit': limit,
        'fields': fields,
    }
    endpoint = '/api/usergroups/'
    headers = {'X-MSTR-ProjectID': None}
    future = future_session.get(endpoint=endpoint, headers=headers, params=params)
    return future


@ErrorHandler(err_msg="Error creating a new user group.")
def create_user_group(connection, body):
    """Create a new user group. The response includes the user group ID, which
    other endpoints use as a request parameter to specify the user group to
    perform an action on. You obtain the authorization token needed to execute
    the request using POST /auth/login; you pass the authorization token in the
    request header. You provide the information to create the.

    User group in the body parameter of the request - including the name and
    description of the user group, the IDs of parent user groups, and the IDs
    of child users or user groups.

    Args:
        connection: Strategy One REST API connection object
        body(JSON):{"name": "string",
                    "description": "string",
                    "memberships": ["string"],
                    "members": ["string"]}
    Returns:
        Complete HTTP response object
    """
    return connection.post(endpoint='/api/usergroups/', json=body)


@ErrorHandler(err_msg="Error getting user group information for a group with ID {id}")
def get_user_group_info(connection, id, error_msg=None):
    """Get information for a specific user group.

    Args:
        connection: Strategy One REST API connection object
        id (string): ID of user group containing your required privileges
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """

    return connection.get(
        endpoint=f'/api/usergroups/{id}',
        headers={
            'X-MSTR-ProjectID': None,
        },
        params={
            'fields': 'id,name,type,subtype,ext_type,abbreviation,dateCreated,'
            'dateModified,version,owner,ancestors,acg,acl,ldapdn',
        },
        # we do not want to get information about members from
        # this endpoint as this is not enough to differentiate between users
        # and user groups
    )


@ErrorHandler(
    err_msg="Error getting user group security roles for a group with ID {id}"
)
def get_security_roles(connection, id, project_id=None, error_msg=None):
    """Get security roles for a specific user group in a specific project.

    Args:
        connection: Strategy One REST API connection object
        id (string): ID of user group containing your required privileges
        project_id (string, optional): Project id string
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """

    return connection.get(
        endpoint=f'/api/usergroups/{id}/securityRoles',
        headers={'X-MSTR-ProjectID': None},
        params={'projectId': project_id},
    )


@ErrorHandler(err_msg="Error getting security filters for user with ID {id}.")
def get_security_filters(
    connection: 'Connection',
    id: str,
    projects: str | list[str] | None = None,
    offset: int = 0,
    limit: int = -1,
    error_msg: str | None = None,
):
    """Get each project level security filter and its corresponding inherited
    security filters for the user group with given ID.

    Args:
        connection: Strategy One REST API connection object
        id (string): User group ID
        projects (str or list of str, optional): collection of projects' ids
            which is used for filtering data
        offset (int, optional): Starting point within the collection of returned
            results. Used to control paging behavior. Default is 0.
        limit (int, optional): Maximum number of items returned for a single
            request. Used to control paging behavior. Use -1 for no limit.
            Default is -1.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    endpoint = f'/api/usergroups/{id}/securityFilters'

    if projects and isinstance(projects, list):
        projects = ','.join(projects)

    params = {'projects.id': projects, 'offset': offset, 'limit': limit}
    return connection.get(endpoint=endpoint, params=params)



from enum import auto

from mstrio.distribution_services import Device
from mstrio.distribution_services.subscription.common import RefreshPolicy
from mstrio.modeling import Attribute, AttributeForm
from mstrio.utils.enum_helper import AutoName, AutoUpperName
from mstrio.utils.helper import Dictable


class Content(Dictable):
    """Class representation of subscription content object.

    Attributes:
        id: Content identifier
        type: Type of content
        name: Name of content
        personalization: Properties object which personalize content
        refresh_condition: RefreshCondition object which sets refresh condition
            of the content

    """

    class Type(AutoName):
        REPORT = auto()
        DOCUMENT = auto()
        CUBE = auto()
        UNSUPPORTED = auto()
        DASHBOARD = 'dossier'

    class Properties(Dictable):
        """Class representation of personalization properties

        Attributes:
            format_mode: Type that defines how much of the Report Services
                document, which has group by units, should be delivered
            view_mode: Indicates the view mode that is supported by this format
            format_type: Content format mode
            export_to_pdf_settings: ExportToPdfSettings object which specifies
                pdf settings if format_type is set to PDF
            delimiter: Format delimiter
            bursting: Bursting settings object
            prompt: Prompt settings object
            file_name: file name of content
        """

        class FormatMode(AutoUpperName):
            DEFAULT = auto()
            CURRENT_PAGE = auto()
            ALL_PAGES = auto()
            ALL_PAGES_SEPARATED = auto()
            CURRENT_WINDOW = auto()
            CURRENT_LAYOUT = auto()

        class ViewMode(AutoUpperName):
            DEFAULT = auto()
            GRID = auto()
            GRAPH = auto()
            BOTH = auto()
            NONINTERACTIVE = auto()

        class FormatType(AutoUpperName):
            PLAIN_TEXT = auto()
            EXCEL = auto()
            HTML = auto()
            PDF = auto()
            STREAMING = auto()
            SWF_MHT = auto()
            SWF_HTML = auto()
            CSV = auto()
            VIEW = auto()
            INTERACTIVE = auto()
            EDITABLE = auto()
            EXPORT_FLASH = auto()
            PHONE = auto()
            TABLET = auto()
            JSON = auto()
            MSTR = auto()

        class ExportToPdfSettings(Dictable):
            """Export to PDF Settings

            Attributes:
                page_option: Specifies whether to export current page or all
                    pages
                page_size: Page size of the PDF file
                orientation: Page orientation. If 'auto' is being used, the
                    exporting engine might swap the width and the height of the
                    page to optimize the export
                page_detail_level: Specifies the detail level of page to
                    be displayed
                include_header: Specifies whether to include header
                include_footer: Specifies whether to include footer
                include_toc: Specifies whether to include table of contents
                filter_summary: Specifies the options of including filter
                    summary in the exported PDF. If 'NONE' is being used, no
                    filter summary will be displayed. If 'BAR' is being used,
                    filter summary bar will be displayed on each page. If 'PAGE'
                    is being used, filter summary page will be displayed at the
                    end of each chapter. If 'ALL' is being used, both filter
                    summary bar and page will be displayed.
                fit_to_page: Specifies whether to fit grid to page
                repeat_column_header: Specifies whether to repeat grid column
                    header
                grid_paging_mode: Specifies how grids should be paginated
            """

            class PageOption(AutoUpperName):
                DEFAULT = auto()
                ALL = auto()
                CURRENT = auto()
                PAGE = auto()

            class PageSize(AutoName):
                A0 = auto()
                A1 = auto()
                A2 = auto()
                A3 = auto()
                A4 = auto()
                A5 = auto()
                B4 = auto()
                B5 = auto()
                LETTER = auto()
                LEGAL = auto()
                LEDGER = auto()
                EXECUTIVE = auto()
                FOLIO = auto()
                STATEMENT = auto()
                UNSUPPORTED = auto()

            class PageDetailLevel(AutoName):
                OVERVIEW = auto()
                DETAILED_PAGES = auto()
                OVERVIEW_AND_DETAILED_PAGES = auto()

            class Orientation(AutoName):
                AUTOMATIC = auto()
                PORTRAIT = auto()
                LANDSCAPE = auto()

            class GridPagingMode(AutoName):
                NONE = auto()
                ENLARGE = auto()

            class FilterSummary(AutoUpperName):
                NONE = auto()
                BAR = auto()
                PAGE = auto()
                ALL = auto()

            # ExportToPdfSettings
            def __init__(
                self,
                page_option: PageOption = PageOption.PAGE,
                page_size: PageSize = PageSize.LETTER,
                orientation: Orientation = Orientation.AUTOMATIC,
                page_detail_level=PageDetailLevel.OVERVIEW,
                include_header: bool = True,
                include_footer: bool = True,
                include_toc: bool = False,
                filter_summary: FilterSummary = FilterSummary.BAR,
                fit_to_page: bool = False,
                repeat_column_header: bool = False,
                grid_paging_mode: GridPagingMode = GridPagingMode.NONE,
            ):
                self.page_option = page_option
                self.page_size = page_size
                self.page_detail_level = page_detail_level
                self.orientation = orientation
                self.include_header = include_header
                self.include_footer = include_footer
                self.include_toc = include_toc
                self.filter_summary = filter_summary
                self.fit_to_page = fit_to_page
                self.repeat_column_header = repeat_column_header
                self.grid_paging_mode = grid_paging_mode

            _FROM_DICT_MAP = {
                "page_option": PageOption,
                "page_size": PageSize,
                "orientation": Orientation,
                "filter_summary": FilterSummary,
            }

        class Bursting(Dictable):
            """Bursting settings

            Attributes:
                slicing_attributes: The list of attributes to slice on
                address_attribute_id: Attribute ID in the email burst feature
                device_id: Device ID in the email burst feature
                form_id: Form ID in the email burst feature
            """

            @staticmethod
            def _create_and_validate_slicing_attr(
                slicing_attributes: (
                    str | Attribute | list[str] | list[Attribute] | None
                ) = None,
                address_attribute_id: str | None = None,
                address_attribute_name: str | None = None,
                address_attribute: Attribute | None = None,
            ) -> list[str]:
                slicing_attributes = [
                    f'{sa.id}~{sa.name}' if isinstance(sa, Attribute) else sa
                    for sa in (
                        slicing_attributes
                        if isinstance(slicing_attributes, list)
                        else [slicing_attributes] if slicing_attributes else []
                    )
                ]

                if not slicing_attributes:
                    if address_attribute:
                        slicing_attributes = [
                            f'{address_attribute.id}~{address_attribute.name}'
                        ]
                    elif address_attribute_id and address_attribute_name:
                        slicing_attributes = [
                            f'{address_attribute_id}~{address_attribute_name}'
                        ]
                    else:
                        msg = (
                            "Slicing attributes must be provided as an `Attribute` "
                            "object or as the `address_attribute_id` and "
                            "`address_attribute_name` arguments."
                        )
                        raise ValueError(msg)

                return slicing_attributes

            def __init__(
                self,
                slicing_attributes: (
                    str | Attribute | list[str] | list[Attribute] | None
                ) = None,
                address_attribute_id: str | None = None,
                address_attribute_name: str | None = None,
                address_attribute: Attribute | None = None,
                device_id: str | None = None,
                device: Device | None = None,
                form_id: str | None = None,
                form: AttributeForm | None = None,
            ):
                """Initialize the `Bursting` object.

                Args:
                    slicing_attributes (str | Attribute | list[str] |
                        list[Attribute], optional): The list of attributes to
                        slice on. Optional if `address_attribute` or
                        `address_attribute_id` and `address_attribute_name` are
                        provided and if it will be the only slicing attribute.
                    address_attribute_id (str, optional): Attribute ID in the
                        email burst feature. Optional if `address_attribute`
                        is provided.
                    address_attribute_name (str, optional): Attribute Name in
                        the email burst feature. Optional if `address_attribute`
                        is provided.
                    address_attribute (Attribute, optional): Attribute object in
                        the email burst feature. If provided, the
                        `address_attribute_id` and `address_attribute_name`
                        arguments will be ignored. Optional if
                        `address_attribute_id` and `address_attribute_name` are
                        provided.
                    device_id (str, optional): Device ID in the email burst
                        feature. Optional if `device` is provided.
                    device (Device, optional): Device object in the email burst
                        feature. If provided, the `device_id` argument will be
                        ignored. Optional if `device_id` is provided.
                    form_id (str, optional): AttributeForm ID in the email burst
                        feature. It has to be the form from the
                        `address_attribute` argument. Optional if `form` is
                        provided.
                    form (AttributeForm, optional): AttributeForm object in the
                        email burst feature. It has to be the form from the
                        `address_attribute` argument. If provided, the `form_id`
                        argument will be ignored. Optional if `form_id` is
                        provided.

                Notes:
                    If `slicing_attributes` argument is not provided, the
                    `address_attribute` object or the `address_attribute_id`
                    and `address_attribute_name` arguments will be used to
                    create it.
                """
                self.slicing_attributes = self._create_and_validate_slicing_attr(
                    slicing_attributes,
                    address_attribute_id,
                    address_attribute_name,
                    address_attribute,
                )
                self.address_attribute_id = (
                    address_attribute.id
                    if isinstance(address_attribute, Attribute)
                    else address_attribute_id
                )
                self.device_id = device.id if isinstance(device, Device) else device_id
                self.form_id = form.id if isinstance(form, AttributeForm) else form_id

        class Prompt(Dictable):
            def __init__(self, enabled: bool, instance_id: str = None):
                self.enabled = enabled
                self.instance_id = instance_id

        # Properties
        def __init__(
            self,
            format_mode: FormatMode = FormatMode.DEFAULT,
            view_mode: ViewMode = ViewMode.DEFAULT,
            format_type: FormatType = FormatType.PDF,
            export_to_pdf_settings: ExportToPdfSettings | None = None,
            delimiter: str | None = None,
            bursting: Bursting | None = None,
            prompt: Prompt | None = None,
            file_name: str | None = None,
            content_modes: list[str] | None = None,
            bookmark_ids: list[str] | None = None,
        ):
            self.format_mode = format_mode
            self.view_mode = view_mode
            self.format_type = format_type
            pdf_format = format_type in [self.FormatType.PDF, self.FormatType.PDF.value]
            self.export_to_pdf_settings = export_to_pdf_settings if pdf_format else None
            self.delimiter = delimiter
            self.bursting = bursting
            self.prompt = prompt
            self.file_name = file_name
            self.content_modes = content_modes
            self.bookmark_ids = bookmark_ids

        _FROM_DICT_MAP = {
            "format_mode": FormatMode,
            "view_mode": ViewMode,
            "format_type": FormatType,
            "export_to_pdf_settings": ExportToPdfSettings.from_dict,
            "bursting": Bursting.from_dict,
            "prompt": Prompt.from_dict,
        }

    class RefreshCondition(Dictable):
        """Dataset refresh condition settings

        Attributes:
            tables: List of TableRefreshInfo objects
            dataset_refresh_policy: Default refresh policy for all the tables
                in the dataset. The setting value must be provided if the tables
                setting value is not provided or empty.
            filters: list of SubscriptionFilter objects
        """

        class SubscriptionFilter(Dictable):
            """Subscription filter. The format of the subscription filters are
                exactly the same as the view filters. Please refer to
                https://microstrategy.github.io/rest-api-docs/common-workflows
                /analytics/filter-data/filter-report-and-cube-instances
                /filter-report-and-cube-instances-using-view-filters
                /code-sample#multiple-filters-on-attribute-forms
                for detailed information. But it should be noted that
                subscription filters only support Filter on attribute forms
                and Multiple filters on attribute forms.

            Attributes:
                type: Filter type
                expression: Metric limits
            """

            def __init__(self, type: str, expression):
                self.type = type
                self.expression = expression

        class TableRefreshInfo(Dictable):
            """Table refresh settings. When the setting is absent in the
                response or the array is empty, the datasetRefreshPolicy setting
                value must be provided and will be used as the refresh policy
                for all the tables, including the new added table after
                the scheduler is set up.

            Attributes:
                id: Table ID
                refresh_policy: Table refresh policy
                alternateSource: AlternateSource object
            """

            class AlternateSource(Dictable):
                """Alternate source

                Attributes:
                    db_role_id: Database instance ID of alternate source
                    namespace: Database namespace of alternate source
                    table_name: Table name of alternate source
                    url: URL of alternate source
                """

                # XXX: Should all of those be optional or all required or what?
                def __init__(
                    self,
                    db_role_id: str | None = None,
                    namespace: str | None = None,
                    table_name: str | None = None,
                    url: str | None = None,
                ):
                    self.db_role_id = db_role_id
                    self.namespace = namespace
                    self.table_name = table_name
                    self.url = url

            # TableRefreshInfo
            def __init__(
                self,
                id: str,
                refresh_policy: RefreshPolicy,
                alternate_source: AlternateSource | None = None,
            ):
                self.id = id
                self.refresh_policy = refresh_policy
                self.alternate_source = alternate_source

            _FROM_DICT_MAP = {
                "refresh_policy": RefreshPolicy,
                "alternate_source": AlternateSource.from_dict,
            }

        # RefreshCondition
        def __init__(
            self,
            tables: list[TableRefreshInfo],
            dataset_refresh_policy: RefreshPolicy | None = None,
            filters: list[SubscriptionFilter] | None = None,
        ):
            self.tables = tables
            self.dataset_refresh_policy = dataset_refresh_policy
            self.filters = filters

        _FROM_DICT_MAP = {
            "dataset_refresh_policy": RefreshPolicy,
            "tables": lambda tables, connection: [
                Content.RefreshCondition.TableRefreshInfo.from_dict(t, connection)
                for t in tables
            ],
            "filters": lambda filters, connection: [
                Content.RefreshCondition.SubscriptionFilter.from_dict(f, connection)
                for f in filters
            ],
        }

    # Content
    def __init__(
        self,
        id: str,
        type: Type | None = None,
        name: str | None = None,
        personalization: Properties | None = None,
        refresh_condition: RefreshCondition | None = None,
    ):
        self.id = id
        self.type = type
        self.name = name
        self.personalization = personalization
        self.refresh_condition = refresh_condition

    _FROM_DICT_MAP = {
        "type": Type,
        "personalization": Properties.from_dict,
        "refresh_condition": RefreshCondition.from_dict,
    }



"""Create a mobile subscription for a selected dashboard and schedule.
1. Connect to the environment using data from workstation
2. List Dashboards and select one that will be used as Content for the subscription
3. List Schedules and select one that will be used for the subscription
4. Define the user that will receive the subscription
5. Create Mobile Subscription with expiration date
6. Execute created mobile subscription
"""

from mstrio.connection import get_connection
from mstrio.distribution_services.schedule.schedule import list_schedules
from mstrio.distribution_services.subscription import Content, MobileSubscription
from mstrio.project_objects import list_dashboards
from mstrio.users_and_groups import list_users

PROJECT_NAME = 'MicroStrategy Tutorial'  # Insert name of project here

conn = get_connection(workstationData, PROJECT_NAME)

# List Dashboards and select one that will be used as Content for the subscription
dashboards = list_dashboards(conn)
selected_dashboard = dashboards[0]

# List Schedules and select one that will be used for the subscription
# It will be 'At Close of Business (Weekday)' in our example here
schedules = list_schedules(conn)
selected_schedule = schedules[1]
print(selected_schedule)

# Define the user that will receive the subscription
# It will be 'Administrator' in our example here
users = list_users(conn, 'Administrator')
selected_user = users[0]
print(selected_user)

# Create Mobile Subscription
mobile_sub = MobileSubscription.create(
    connection=conn,
    name='<Name of the subscription>',
    schedules=[selected_schedule],
    contents=Content(
        id=selected_dashboard.id,
        type=Content.Type.DASHBOARD,
        personalization=Content.Properties(
            format_type=Content.Properties.FormatType.HTML
        ),
    ),
    recipients=[selected_user.id],
    delivery_expiration_date='2024-12-31',
    delivery_expiration_timezone='Europe/London',
)

# Execute the subscription
mobile_sub.execute()



from dataclasses import asdict, dataclass
from enum import Enum, auto
from typing import TYPE_CHECKING

from humps import camelize, decamelize
from stringcase import capitalcase

from mstrio.utils.enum_helper import AutoName
from mstrio.utils.helper import Dictable, delete_none_values, snake_to_camel

if TYPE_CHECKING:
    from mstrio.connection import Connection


class DynamicDateTimeType(AutoName):
    """Enumeration constant indicating type of dynamic date time object"""

    DATE = auto()
    TIME = auto()
    DATE_TIME = auto()


class DateMode(AutoName):
    """Enumeration constant indicating mode of date"""

    DYNAMIC = auto()
    STATIC = auto()


class HourMode(AutoName):
    """Enumeration constant indicating mode of hour"""

    DYNAMIC = auto()
    STATIC = auto()


class MinuteAndSecondMode(AutoName):
    """Enumeration constant indicating mode of minute and second"""

    DYNAMIC = auto()
    STATIC = auto()


class DayOfWeek(Enum):
    """Enumeration used to indicate day of week"""

    SUNDAY = 0
    MONDAY = 1
    TUESDAY = 2
    WEDNESDAY = 3
    THURSDAY = 4
    FRIDAY = 5
    SATURDAY = 6


class VersatileDate(Dictable):
    """Base class for classes representing dynamic or static date objects"""

    _MODE = None

    @staticmethod
    def dispatch(
        source, connection: 'Connection | None' = None
    ) -> type['VersatileDate']:
        """Returns an appropriate VersatileDate type object from the
            provided source

        Args:
            source: object that specifies the VersatileDate object that
                will be returned
            connection (optional): Strategy One connection object returned
            by `connection.Connection()`"""
        mode = DateMode(source.get('mode'))
        if mode == DateMode.DYNAMIC:
            cls = DynamicVersatileDate
        elif mode == DateMode.STATIC:
            cls = StaticVersatileDate

        return cls.from_dict(source, connection)


class DateAdjustment(Dictable):
    """Base class for date adjustment objects"""

    pass


@dataclass(frozen=True)
class AdjustmentNone(DateAdjustment):
    """Indicate that there is no further adjustments"""

    pass


@dataclass
class AdjustmentWeeklyByDayOfWeek(DateAdjustment):
    """Adjustment at the week level by day of week

    Attributes:
        day_of_week: day of week, value of DayOfWeek enum
    """

    _FROM_DICT_MAP = {'day_of_week': DayOfWeek}
    day_of_week: DayOfWeek


@dataclass
class AdjustmentMonthlyByDay(DateAdjustment):
    """Adjustment at the month level by day

    Attributes:
        day: day of month (1 - 31)
    """

    day: int


@dataclass
class AdjustmentMonthlyByDayOfWeek(DateAdjustment):
    """Adjustment at the month level by day of week

    Attributes:
        week_ordinal: number indicating 1st, 2nd, 3rd, etc.
        day_of_week: day of week, value of DayOfWeek enum
    """

    _FROM_DICT_MAP = {'day_of_week': DayOfWeek}
    week_ordinal: int
    day_of_week: DayOfWeek


@dataclass
class AdjustmentMonthlyByReverseCount(DateAdjustment):
    """Adjustment at the month level by reverse count

    Attributes:
        days: number of days to reverse count
        months: number of months to reverse count
    """

    days: int
    months: int


@dataclass
class AdjustmentQuarterlyByDay(DateAdjustment):
    """Adjustment at the quarter level by day

    Attributes:
        day: day in the quarter
    """

    day: int


@dataclass
class AdjustmentQuarterlyByDayOfWeek(DateAdjustment):
    """Adjustment at the quarter level by day of week

    Attributes:
        month_ordinal: number indicating 1st, 2nd, or 3rd month in the quarter
        week_ordinal: number indicating 1st, 2nd, 3rd, etc.
        day_of_week: day of week, value of DayOfWeek enum
    """

    _FROM_DICT_MAP = {'day_of_week': DayOfWeek}
    month_ordinal: int
    week_ordinal: int
    day_of_week: DayOfWeek


@dataclass
class AdjustmentQuarterlyByReverseCount(DateAdjustment):
    """Adjustment at the quarter level by reverse count

    Attributes:
        days: number of days to reverse count
    """

    days: int


@dataclass
class AdjustmentYearlyByDate(DateAdjustment):
    """Adjustment at the year level by date

    Attributes:
        month: Month of the year (1 - 12)
        day: Day of month (1 - 31)
    """

    month: int
    day: int


@dataclass
class AdjustmentYearlyByDayOfWeek(DateAdjustment):
    """Adjustment at the year level by day of week

    Attributes:
        month: Month of the year (1 - 12)
        week_ordinal: Number indicating 1st, 2nd, 3rd, etc.
        day_of_week: day of week, value of DayOfWeek enum
    """

    _FROM_DICT_MAP = {'day_of_week': DayOfWeek}
    month: int
    week_ordinal: int
    day_of_week: DayOfWeek


@dataclass
class DynamicVersatileDate(VersatileDate):
    """Object that represents a dynamic date.
    Its various settings define a function that will resolve to a date using
    the current date as the input.

    Attributes:
        day_offset: The day offset
        month_offset: The month offset
        exclude_weekends: Whether weekend days should be excluded during
            calculation, default: False
        adjustment: adjustment of a date, instance of one of classes inheriting
            from DateAdjustment class
    """

    _MODE = DateMode.DYNAMIC
    _ALLOW_NONE_ATTRIBUTES = ['adjustment_none']
    day_offset: int
    month_offset: int
    adjustment: DateAdjustment = AdjustmentNone()
    exclude_weekends: bool | None = None

    def to_dict(self, camel_case: bool = True) -> dict:
        result = asdict(self)
        result['mode'] = self._MODE.value

        # every type of adjustment has its own unique attribute name
        # and only one can be present in json
        # so generate it based on class name
        attr_name = decamelize(self.adjustment.__class__.__name__)
        result[attr_name] = result.pop('adjustment')

        result = delete_none_values(
            result, recursion=False, whitelist_attributes=self._ALLOW_NONE_ATTRIBUTES
        )
        return snake_to_camel(result) if camel_case else result

    @classmethod
    def from_dict(
        cls, source: dict, connection: 'Connection | None' = None, to_snake_case=True
    ) -> 'DynamicVersatileDate':
        try:
            key, value = next(
                (key, value)
                for key, value in source.items()
                if key.startswith('adjustment')
            )
        except StopIteration:
            raise AttributeError(
                "Missing field for adjustment of a date in source data."
            )

        cls_name = capitalcase(camelize(key))
        cls_obj = globals()[cls_name]
        adjustment = cls_obj(**value)

        result = super().from_dict(source, connection, to_snake_case)
        result.adjustment = adjustment

        return result


@dataclass
class StaticVersatileDate(VersatileDate):
    """Object that represents a static date

    Attributes:
        value: ISO 8601 string of the date component
    """

    _MODE = DateMode.STATIC
    value: str

    def to_dict(self, camel_case: bool = True) -> dict:
        result = asdict(self)
        result['mode'] = self._MODE.value

        return snake_to_camel(result) if camel_case else result


@dataclass
class VersatileTime(Dictable):
    """Object that represents either a dynamic time or a static time

    Attributes:
        hour_mode: The hour component (0-23). Used when the hour is static.
        hour_offset: The offset from the current hour. Used when the hour
            is dynamic.
        minute_and_second_mode: The minute component (0-59). Used when
            the minute is static.
        minute_offset: The offset from the current minute. Used when
            the minute is dynamic.
        second: The second component (0-59). Used when the second is static.
        second_offset: The offset from the current second. Used when the second
            is dynamic.
    """

    _FROM_DICT_MAP = {
        'hour_mode': HourMode,
        'minute_and_second_mode': MinuteAndSecondMode,
    }
    hour_mode: HourMode
    minute_and_second_mode: MinuteAndSecondMode
    hour: int | None = None
    hour_offset: int | None = None
    minute: int | None = None
    minute_offset: int | None = None
    second: int | None = None
    second_offset: int | None = None


@dataclass
class DynamicDateTimeStructure(Dictable):
    """Object that represents a date/time. It can be date-only, or time-only,
    or date-time. The date part and the time part can be dynamic or static
    independently. Typically, this object is used when at least one of the two
    parts is dynamic. Although it's capable of describing a static date/time,
    the much simpler Variant is preferable in that case.

    Attributes:
        type: represents what part is being used, either date, time or date/time
        date: Object that represents either a dynamic date or a static date
        time: Object that represents either a dynamic time or a static time
    """

    _FROM_DICT_MAP = {
        'type': DynamicDateTimeType,
        'date': VersatileDate.dispatch,
        'time': VersatileTime,
    }
    type: DynamicDateTimeType
    date: VersatileDate | None = None
    time: VersatileTime | None = None



from requests import Response

from mstrio.connection import Connection
from mstrio.utils.error_handlers import ErrorHandler


@ErrorHandler(err_msg="Error getting color palettes.")
def list_palettes(
    connection: Connection,
    project_id: str | None = None,
    fields: str | None = None,
    error_msg: str | None = None,
) -> Response:
    """Get a list of color palettes.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        project_id (str, optional): Project ID.
        fields (str, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model.
        error_msg (str, optional): Custom error message for error handling

    Returns:
        Complete HTTP response object. 200 on success.
    """
    return connection.get(
        endpoint='/api/palettes',
        headers={'X-MSTR-ProjectID': project_id},
        params={
            'fields': fields,
        },
    )


@ErrorHandler(err_msg="Error creating the color palette.")
def create_palette(
    connection: Connection,
    body: dict,
    project_id: str | None = None,
    error_msg: str | None = None,
) -> Response:
    """Create a new color palette.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        project_id (str, optional): Project ID.
        body (dict): JSON-formatted body describing the new palette
        error_msg (str, optional): Custom error message for error handling

    Returns:
        HTTP response object returned by the Strategy One REST server.
    """
    return connection.post(
        endpoint='/api/palettes',
        json=body,
        headers={'X-MSTR-ProjectID': project_id},
    )


@ErrorHandler(err_msg="Error getting a palette with ID: {id}")
def get_palette(
    connection: Connection,
    id: str,
    project_id: str | None = None,
    error_msg: str | None = None,
) -> Response:
    """Get a color palette by ID.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        project_id (str, optional): Project ID.
        id (str): ID of the palette to get
        error_msg (str, optional): Custom error message for error handling

    Returns:
        HTTP response object returned by the Strategy One REST server.
    """
    return connection.get(
        endpoint=f'/api/palettes/{id}',
        headers={'X-MSTR-ProjectID': project_id},
    )


@ErrorHandler(err_msg="Error updating the palette with ID: {id}")
def update_palette(
    connection: Connection,
    id: str,
    body: dict,
    project_id: str | None = None,
    error_msg: str | None = None,
) -> Response:
    """Update an existing color palette.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        project_id (str, optional): Project ID.
        id (str): ID of the palette to update
        body (dict): JSON-formatted body describing the updated palette
        error_msg (str, optional): Custom error message for error handling

    Returns:
        HTTP response object returned by the Strategy One REST server.
    """
    return connection.put(
        endpoint=f'/api/palettes/{id}',
        json=body,
        headers={'X-MSTR-ProjectID': project_id},
    )



"""
Duplicate user subscriptions across all projects.

1. Connect to the environment using data from workstation
2. Get user with the given id
2. Get all projects that the authenticated user has access to
4. For each project get all subscriptions of the user which was earlier
   retrieved
5. Duplicate each received subscription and change its name by adding " Copy" at
   the end and specify new owner by providing some other id
"""

from mstrio.api.projects import get_projects
from mstrio.api.subscriptions import create_subscription
from mstrio.connection import Connection, get_connection
from mstrio.distribution_services import list_subscriptions
from mstrio.users_and_groups import User


def duplicate_user_subscriptions(connection: Connection, owner: str | User) -> None:
    """Duplicate user subscriptions across all projects.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        owner (str, User): The owner of the subscriptions to duplicate.
    """

    # get all projects that the authenticated user has access to
    response = get_projects(connection, whitelist=[('ERR014', 403)])
    projects = response.json() if response.ok else []

    owner_id = owner.id if isinstance(owner, User) else owner

    for project in projects:
        project_id = project['id']
        subscriptions = list_subscriptions(
            connection=connection,
            project_id=project_id,
            owner={'id': owner_id},
        )
        for subscription in subscriptions:
            subscription_dict = subscription.to_dict()
            # Update subscription name and owner
            subscription_dict.update(
                name=f'{subscription_dict["name"]} Copy',
                owner={'id': '51F189429A439FAA734550A753978285'},
            )
            create_subscription(
                connection, project_id=project_id, body=subscription_dict
            )


# connect to environment without providing user credentials
# variable `workstationData` is stored within Workstation
conn = get_connection(workstationData, 'MicroStrategy Tutorial')

user = User(connection=conn, id='7FC05A65473CE2FD845CE6A1D3F13233')
duplicate_user_subscriptions(conn, user)



from requests import Response

from mstrio.connection import Connection
from mstrio.object_management import Object, full_search
from mstrio.utils.error_handlers import ErrorHandler


def list_scripts(
    connection: Connection, project_id: str, to_dictionary=True
) -> list[dict] | list[Object]:
    """Get a list of scripts.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        project_id (str): ID of the project
        to_dictionary (bool, optional): If True returns dict (default),
            otherwise returns Object.

    Returns:
        List of scripts.
    """
    return full_search(
        connection, project=project_id, object_types=76, to_dictionary=to_dictionary
    )


@ErrorHandler(err_msg="Error creating Script.")
def create_script(
    connection: Connection, project_id: str, body: dict, error_msg: str | None = None
) -> Response:
    """Create a script.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        body (dict): JSON-formatted body of the new script
        project_id (str): ID of the project
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object returned by the Strategy One REST server.
    """
    return connection.post(
        endpoint='/api/scripts',
        headers={'X-MSTR-ProjectID': project_id},
        json=body,
    )


@ErrorHandler(err_msg="Error deleting Script with ID: {id}")
def delete_script(
    connection: Connection, id: str, project_id: str, error_msg: str | None = None
) -> Response:
    """Delete a script.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        id (str): ID of the script to be deleted
        project_id (str): ID of the project
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object returned by the Strategy One REST server.
    """
    return connection.delete(
        endpoint=f'/api/scripts/{id}', headers={'X-MSTR-ProjectID': project_id}
    )



"""This is the demo script to show how to manage fences.
Its basic goal is to present what can be done with this module and to ease
its usage.
"""

from mstrio.connection import get_connection
from mstrio.server.fence import Fence, FenceType, list_fences
from mstrio.server.project import Project
from mstrio.users_and_groups import User, UserGroup

# Define a variable which can be later used in a script
PROJECT_ID = $project_id

conn = get_connection(workstationData, project_id=PROJECT_ID)

# List fences
# Note: No Fences exist in a default environment
fences = list_fences(connection=conn)
fences_as_dicts = list_fences(connection=conn, to_dictionary=True)

# Define a variable which can be later used in a script
RANK = $rank
FENCE_NAME = $fence_name
NODE_NAME = $node_name
USER_ID = $user_id
USER_GROUP_ID = $user_group_id

# Create a Fence
fence = Fence.create(
    connection=conn,
    rank=RANK,
    name=FENCE_NAME,
    type=FenceType.USER_FENCE,
    nodes=[NODE_NAME],
    users=[User(connection=conn, id=USER_ID)],
    user_groups=[UserGroup(connection=conn, id=USER_GROUP_ID)],
    projects=[Project(connection=conn, id=PROJECT_ID)],
)

# Get a Fence by it's id or name
fence = Fence(connection=conn, id=fence.id)
fence = Fence(connection=conn, name=FENCE_NAME)

# Define variables which can be later used in a script
RANK_2 = $rank_2
USER_ID_2 = $user_id_2
USER_GROUP_ID_2 = $user_group_id_2
PROJECT_ID_2 = $project_id_2

# Alter a Fence
fence.alter(
    rank=RANK_2,
    users=[User(connection=conn, id=USER_ID_2)],
    user_groups=[UserGroup(connection=conn, id=USER_GROUP_ID_2)],
    projects=[Project(connection=conn, id=PROJECT_ID_2)],
)

# Delete a Fence
fence.delete(force=True)



def convert_settings_to_byte(settings: dict, mapping: dict) -> dict:
    def convert(setting, value):
        unit = mapping.get(setting)
        if unit is not None:
            if unit == 'B':
                return value * (1024**2)
            elif unit == 'KB':
                return value * 1024
        return value

    return {setting: convert(setting, value) for setting, value in settings.items()}


def convert_settings_to_mega_byte(settings: dict, mapping: dict) -> dict:
    def convert(setting, value):
        unit = mapping.get(setting)
        if unit is not None:
            if unit == 'B':
                return value // (1024**2)
            elif unit == 'KB':
                return value // 1024
        return value

    return {setting: convert(setting, value) for setting, value in settings.items()}



from dataclasses import dataclass
from typing import Any

from mstrio.utils.helper import Dictable, deprecation_warning


deprecation_warning("mstrio.project_objects.prompt", "mstrio.modeling.prompt", "11.6.9")


@dataclass
class Prompt(Dictable):
    """A Strategy One class representing a prompt.

    Attributes:
        type (str): Type of the prompt
            Possible values are:
                - UNSUPPORTED
                - VALUE
                - ELEMENTS
                - EXPRESSION
                - OBJECTS
                - LEVEL
        answers (Any | list[Any]): Singular answer or list of
            answers to the prompt.
        key (str, optional): Unique key of the prompt.
        id (str, optional): ID of the prompt.
        name (str, optional): Name of the prompt.
        use_default (bool, optional): Whether to use default value.
            If True, provided answer will be ignored. Defaults to False.

    Note that only one of the key, id, or name needs to be provided.
    It is recommended to always provide the key as it's always unique
    as opposed to ID or name that can be shared among prompts.
    """

    type: str
    answers: Any | list[Any]
    key: str | None = None
    id: str | None = None
    name: str | None = None
    use_default: bool = False



import logging
from enum import auto
from typing import TYPE_CHECKING

import pandas as pd

from mstrio import config
from mstrio.api import incremental_refresh_reports as refresh_api
from mstrio.api import reports as reports_api
from mstrio.connection import Connection
from mstrio.modeling import (
    Expression,
    ExpressionFormat,
    ObjectSubType,
    Prompt,
    SchemaObjectReference,
)
from mstrio.object_management import Folder, SearchPattern, full_search
from mstrio.object_management.folder import get_folder_id_from_path
from mstrio.project_objects import OlapCube
from mstrio.project_objects.datasets.helpers import AdvancedProperties, Template
from mstrio.server import Job
from mstrio.types import ObjectSubTypes, ObjectTypes
from mstrio.users_and_groups.user import User
from mstrio.utils.entity import CopyMixin, DeleteMixin, Entity, MoveMixin
from mstrio.utils.enum_helper import AutoName, get_enum_val
from mstrio.utils.helper import (
    delete_none_values,
    filter_params_for_func,
    find_object_with_name,
    get_objects_id,
)
from mstrio.utils.parser import Parser
from mstrio.utils.resolvers import get_project_id_from_params_set
from mstrio.utils.response_processors import objects as objects_processors
from mstrio.utils.version_helper import class_version_handler, method_version_handler
from mstrio.utils.vldb_mixin import ModelVldbMixin

if TYPE_CHECKING:
    from mstrio.server.project import Project

logger = logging.getLogger(__name__)


@method_version_handler('11.3.0600')
def list_incremental_refresh_reports(
    connection: Connection,
    name: str | None = None,
    pattern: SearchPattern | int = SearchPattern.CONTAINS,
    project: 'Project | str | None' = None,
    project_id: str | None = None,
    project_name: str | None = None,
    to_dictionary: bool = False,
    limit: int | None = None,
    folder_id: str | None = None,
    folder_path: str | None = None,
    show_expression_as: ExpressionFormat | str = ExpressionFormat.TREE,
    show_filter_tokens: bool = False,
    show_advanced_properties: bool = True,
    **filters,
) -> list["IncrementalRefreshReport"] | list[dict]:
    """Get a list of Incremental Refresh Reports.
    Optionally filter reports by specifying 'name'.

    Optionally use `to_dictionary` to choose output format.

    Wildcards available for 'name':
        ? - any character
        * - 0 or more of any characters
        e.g. name = ?onny will return Sonny and Tonny

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        name (string, optional): value the search pattern is set to, which
            will be applied to the names of reports being searched
        pattern (SearchPattern enum or int, optional): pattern to search
            for, such as Begin With or Contains. Possible values are available
            in ENUM mstrio.object_management.SearchPattern.
            Default value is BEGIN WITH (4).
        project (Project | str, optional): Project object or ID or name
            specifying the project. May be used instead of `project_id` or
            `project_name`.
        project_id (str, optional): Project ID
        project_name (str, optional): Project name
        to_dictionary (bool, optional): If True returns dict, by default (False)
            returns Report objects
        limit (integer, optional): limit the number of elements returned. If
            None all object are returned.
        folder_id (string, optional): ID of a folder where the search
            will be performed. Defaults to None.
        folder_path (str, optional): Path of the folder in which to look for
            reports. Can be provided as an alternative to the `folder_id`
            parameter. If both are provided, `folder_id` is used.
            The path has to be provided in the following format:
                if it's inside of a project, example:
                    /MicroStrategy Tutorial/Public Objects
                if it's a root folder, example:
                    /CASTOR_SERVER_CONFIGURATION/Users
        show_expression_as (ExpressionFormat or str, optional): specify how
            expressions should be presented
            Available values:
            - None
            (expression is returned in "text" format)
            - `ExpressionFormat.TREE` or `tree`
            (expression is returned in `text` and `tree` formats)
            - `ExpressionFormat.TOKENS` or `tokens`
            (expression is returned in `text` and `tokens` formats)
        show_filter_tokens (bool, optional): Specify whether "qualification"
            is returned in "tokens" format,
            along with `text` and `tree` formats.
            - If omitted or false, only `text` and `tree`
            formats are returned.
            - If true, all `text`, `tree` and `tokens` formats are returned.
        show_advanced_properties (bool, optional): Specify whether to retrieve
            the values of the advanced properties. If false, nothing will be
            returned for the advanced properties, default True.
        filters: Available filter parameters: ['id', 'name', 'type',
            'subtype', 'date_created', 'date_modified', 'version', 'owner',
            'ext_type', 'view_media', 'certified_info']
    """
    proj_id = get_project_id_from_params_set(
        connection,
        project,
        project_id,
        project_name,
    )
    filters = filters or {}

    if folder_path and not folder_id:
        folder_id = get_folder_id_from_path(connection=connection, path=folder_path)
    if folder_id:
        filters = filters | {'root': folder_id}

    objects = full_search(
        connection,
        object_types=ObjectSubTypes.INCREMENTAL_REFRESH_REPORT,
        project=proj_id,
        name=name,
        pattern=pattern,
        limit=limit,
        **filters,
    )

    if to_dictionary:
        return objects

    return [
        IncrementalRefreshReport.from_dict(
            obj
            | {
                'show_expression_as': (
                    show_expression_as
                    if isinstance(show_expression_as, ExpressionFormat)
                    else ExpressionFormat(show_expression_as)
                ),
                'show_filter_tokens': show_filter_tokens,
                'show_advanced_properties': show_advanced_properties,
            },
            connection,
        )
        for obj in objects
    ]


@class_version_handler('11.3.0600')
class IncrementalRefreshReport(
    Entity, CopyMixin, MoveMixin, DeleteMixin, ModelVldbMixin
):
    """Python representation of Strategy One Incremental Refresh Report object.

    Attributes:
        name: (str) Name of the Incremental Refresh Report
        id: (str) ID of the Incremental Refresh Report
        description: (str) Description of the Incremental Refresh Report
        destination_folder_id: (str) ID of the folder
        target_cube: (SchemaObjectReference) Reference to an Intelligent Cube
        refresh_type: (str) Mode the report will be refreshed in
        increment_type: (str) Mode the report will be executed in
        template: (Template) Template of information layout across dimensions
        filter: (Expression) Filtering expression for the report
    """

    class IncrementType(AutoName):
        REPORT = auto()
        FILTER = auto()

    class RefreshType(AutoName):
        INSERT = auto()
        UPDATE = auto()
        UPDATE_ONLY = auto()
        REPLACE = auto()
        DELETE = auto()

    _OBJECT_TYPE = ObjectTypes.REPORT_DEFINITION

    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP,
        'target_cube': SchemaObjectReference.from_dict,
        'increment_type': IncrementType,
        'refresh_type': RefreshType,
        'template': Template.from_dict,
        'filter': Expression.from_dict,
        'advanced_properties': AdvancedProperties.from_dict,
        'owner': User.from_dict,
    }

    _API_GETTERS = {
        (
            'type',
            'subtype',
            'ext_type',
            'date_created',
            'date_modified',
            'version',
            'owner',
            'ancestors',
            'acg',
            'acl',
            'hidden',
        ): objects_processors.get_info,
        (
            'id',
            'name',
            'description',
            'sub_type',
            'date_created',
            'date_modified',
            'path',
            'version_id',
            'is_embedded',
            'primary_locale',
            'destination_folder_id',
            'target_cube',
            'refresh_type',
            'increment_type',
            'template',
            'filter',
            'advanced_properties',
        ): refresh_api.get_incremental_refresh_report,
    }
    _API_PATCH: dict = {
        (
            'target_cube',
            'refresh_type',
            'increment_type',
            'template',
            'filter',
            'advanced_properties',
        ): (refresh_api.update_incremental_refresh_report, "put"),
        (
            'name',
            'description',
            'destination_folder_id',
            'is_embedded',
            'folder_id',
            'hidden',
            'owner',
        ): (objects_processors.update, 'partial_put'),
    }
    _PATCH_PATH_TYPES = {
        **Entity._PATCH_PATH_TYPES,
        'destination_folder_id': str,
        'is_embedded': bool,
        'target_cube': dict,
        'refresh_type': str,
        'increment_type': str,
        'filter': dict,
        'advanced_properties': dict,
    }
    _MODEL_VLDB_API = {
        'GET_ADVANCED': refresh_api.get_incremental_refresh_report,
        'PUT_ADVANCED': refresh_api.update_incremental_refresh_report,
        'GET_APPLICABLE': refresh_api.get_incremental_refresh_report_vldb_properties,
    }

    _ALLOW_NONE_ATTRIBUTES = ['filter', 'template']
    _KEEP_CAMEL_CASE = ['vldbProperties', 'metricJoinTypes', 'attributeJoinTypes']

    def __init__(
        self,
        connection: Connection,
        id: str | None = None,
        name: str | None = None,
        show_expression_as: ExpressionFormat | str = ExpressionFormat.TREE,
        show_filter_tokens: bool = False,
        show_advanced_properties: bool = False,
    ):
        """Initialize an Incremental Refresh Report object.

        Args:
            connection: Strategy One connection object returned by
                `connection.Connection()`
            id (str, optional): ID of the Incremental Refresh Report
            name (str, optional): Name of the Incremental Refresh Report
            show_expression_as (ExpressionFormat or str, optional): specify how
                expressions should be presented
                Available values:
                - None
                (expression is returned in "text" format)
                - `ExpressionFormat.TREE` or `tree`
                (expression is returned in `text` and `tree` formats)
                - `ExpressionFormat.TOKENS` or `tokens`
                (expression is returned in `text` and `tokens` formats)
            show_filter_tokens (bool, optional): Specify whether "qualification"
                is returned in "tokens" format,
                along with `text` and `tree` formats.
                - If omitted or false, only `text` and `tree`
                formats are returned.
                - If true, all `text`, "tree" and `tokens` formats are returned.
            show_advanced_properties (bool, optional): Specify whether to
                retrieve the values of the advanced properties.
                The advanced properties are presented in the following groups:
                    "vldbProperties": A list of properties as determined by
                        the common infrastructure.
                    "metricJoinTypes": A list of Metric Join Types, one for each
                        metric that appears in the template.
                    "attributeJoinTypes": A list of Attribute Join Types, one
                        for each attribute that appears in the template.
                If omitted or false, nothing will be returned for the advanced
                    properties.
                If true, all applicable advanced properties are returned.
        """
        connection._validate_project_selected()

        if id is None:
            if name is None:
                raise ValueError(
                    "Please specify either 'name' or 'id' parameter in the constructor."
                )

            reports = find_object_with_name(
                connection=connection,
                cls=self.__class__,
                name=name,
                listing_function=list_incremental_refresh_reports,
            )
            object_id = reports['id']
        else:
            object_id = id

        super().__init__(
            connection=connection,
            object_id=object_id,
            show_expression_as=show_expression_as,
            show_filter_tokens=show_filter_tokens,
            show_advanced_properties=show_advanced_properties,
        )

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)

        self.primary_locale = kwargs.get('primary_locale')
        self.is_embedded = kwargs.get('is_embedded')
        self.destination_folder_id = kwargs.get('destination_folder_id')

        self._sub_type = (
            ObjectSubType(kwargs.get('sub_type')) if kwargs.get('sub_type') else None
        )
        self._path = kwargs.get('path')

        self.target_cube = (
            SchemaObjectReference.from_dict(kwargs.get('target_cube'))
            if kwargs.get('target_cube')
            else None
        )
        self.refresh_type = (
            self.RefreshType(kwargs.get('refresh_type'))
            if kwargs.get('refresh_type')
            else None
        )
        self.increment_type = (
            self.IncrementType(kwargs.get('increment_type'))
            if kwargs.get('increment_type')
            else None
        )
        self.template = (
            Template.from_dict(kwargs.get('template'), self.connection)
            if kwargs.get('template')
            else None
        )
        self.filter = (
            Expression.from_dict(kwargs.get('filter'), self.connection)
            if kwargs.get('filter')
            else None
        )
        self.advanced_properties = (
            AdvancedProperties.from_dict(
                kwargs.get('advanced_properties'), self.connection
            )
            if kwargs.get('advanced_properties')
            else None
        )

        show_expression_as = kwargs.get('show_expression_as', 'tree')
        self._show_expression_as = (
            show_expression_as
            if isinstance(show_expression_as, ExpressionFormat)
            else ExpressionFormat(show_expression_as)
        )
        self._show_filter_tokens = kwargs.get('show_filter_tokens', False)
        self._show_advanced_properties = kwargs.get('show_advanced_properties', False)
        self._prompts = None

    def execute(
        self,
        fields: str | None = None,
        project: 'Project | str | None' = None,
        project_id: str | None = None,
        project_name: str | None = None,
    ) -> Job:
        """Execute (run) the report.

        Args:
            fields (str, optional): A comma-separated list of fields to include
                in the response. By default, all fields are returned.
            project (Project | str, optional): Project object or ID or name
                specifying the project. May be used instead of `project_id` or
                `project_name`.
            project_id (str, optional): Project ID
            project_name (str, optional): Project name

        Returns:
            Job instance.
        """
        proj_id = get_project_id_from_params_set(
            self.connection,
            project,
            project_id,
            project_name,
        )

        response = refresh_api.execute_incremental_refresh_report(
            self.connection,
            id=self.id,
            project_id=proj_id,
            fields=fields,
        )

        job_data = response.json()
        if config.verbose:
            logger.info(
                f"Execution of Incremental Refresh Report: '{self.name}' has been "
                f"successfully scheduled under job: {job_data}."
            )
        return Job.from_dict(job_data, self.connection)

    @classmethod
    def create(
        cls,  # NOSONAR
        connection: Connection,
        name: str,
        destination_folder: Folder | str | None = None,
        destination_folder_path: str | None = None,
        target_cube: OlapCube | SchemaObjectReference | dict | None = None,
        increment_type: IncrementType | str | None = None,
        refresh_type: RefreshType | str | None = None,
        template: Template | dict | None = None,
        filter: Expression | dict | None = None,
        advanced_properties: dict | None = None,
        description: str | None = None,
        primary_locale: str | None = None,
        show_expression_as: ExpressionFormat | str = ExpressionFormat.TREE,
        show_filter_tokens: bool = False,
        show_advanced_properties: bool = False,
    ) -> 'IncrementalRefreshReport':
        """Create a new Incremental Refresh Report.

        Args:
            connection: Strategy One connection object returned by
                `connection.Connection()`
            name (str, optional): Name of the Incremental Refresh Report
            destination_folder (Folder, str): Folder object or folder ID where
                the new object will be saved
            destination_folder_path (str, optional): Path of the folder in which
                to create the report. If both folder and path parameters are
                provided, `destination_folder` is used.
                The path has to be provided in the following format:
                    if it's inside of a project, example:
                        /MicroStrategy Tutorial/Public Objects
                    if it's a root folder, example:
                        /CASTOR_SERVER_CONFIGURATION/Users
            target_cube (OlapCube, SchemaObjectReference, dict, optional):
                Reference to an Intelligent Cube
            increment_type (IncrementType, str, optional): Mode the report will
                be executed in. Accepted values: 'report', 'filter'
            refresh_type (RefreshType, str, optional): Mode the report will be
                refreshed in. Accepted values: 'update', 'insert', 'delete',
                'update_only', 'replace'
            template (Template, dict, optional): Template of information layout
                across dimensions
            filter (Expression, dict, optional): Filtering expression for the
                report. Accepted values: 'tree', 'tokens'
            advanced_properties (dict, optional): Advanced properties of the
                report
            description (str, optional): Description of the MSTR object
            primary_locale (str, optional): The primary locale of the object,
                in the IETF BCP 47 language tag format, such as "en-US"
            show_expression_as (ExpressionFormat or str, optional): specify how
                expressions should be presented
                Available values:
                - None
                (expression is returned in "text" format)
                - `ExpressionFormat.TREE` or `tree`
                (expression is returned in `text` and `tree` formats)
                - `ExpressionFormat.TOKENS` or `tokens`
                (expression is returned in `text` and `tokens` formats)
            show_filter_tokens (bool, optional): Specify whether "qualification"
                is returned in "tokens" format,
                along with `text` and `tree` formats.
                - If omitted or false, only `text` and `tree`
                formats are returned.
                - If true, all `text`, "tree" and `tokens` formats are returned.
        """

        if destination_folder_path and not destination_folder:
            destination_folder = get_folder_id_from_path(
                connection=connection, path=destination_folder_path
            )
        information = {
            'name': name,
            'description': description,
            'destinationFolderId': get_objects_id(destination_folder, Folder),
            'primaryLocale': primary_locale,
        }

        if isinstance(target_cube, SchemaObjectReference):
            target_cube = target_cube.to_dict()
        elif isinstance(target_cube, OlapCube):
            target_cube = SchemaObjectReference(
                object_id=target_cube.id,
                name=target_cube.name,
                sub_type=ObjectSubType.REPORT_CUBE,
            ).to_dict()

        body = {
            'information': delete_none_values(information, recursion=True),
            'targetCube': target_cube,
            'incrementType': get_enum_val(increment_type, cls.IncrementType),
            'refreshType': get_enum_val(refresh_type, cls.RefreshType),
            'template': (
                template.to_dict() if isinstance(template, Template) else template
            ),
            'filter': (
                filter.to_dict() if isinstance(filter, Expression) else (filter or {})
            ),
            'advanced_properties': advanced_properties,
        }

        response = refresh_api.create_incremental_refresh_report(
            connection=connection,
            body=body,
            show_expression_as=get_enum_val(show_expression_as, ExpressionFormat),
            show_filter_tokens=show_filter_tokens,
            show_advanced_properties=show_advanced_properties,
        ).json()

        logger.info(
            f"Successfully created Incremental Refresh Report named: "
            f"'{name}' with ID: '{response['id']}'"
        )

        return cls.from_dict(
            source={
                **response,
                'show_expression_as': show_expression_as,
                'show_filter_tokens': show_filter_tokens,
                'show_advanced_properties': show_advanced_properties,
            },
            connection=connection,
        )

    @classmethod
    def create_from_cube(
        cls,
        connection: Connection,
        name: str,
        destination_folder: Folder | str | None = None,
        destination_folder_path: str | None = None,
        target_cube: OlapCube | SchemaObjectReference | dict | None = None,
        refresh_type: RefreshType | str | None = None,
        filter: Expression | dict | None = None,
        advanced_properties: dict | None = None,
        description: str | None = None,
        primary_locale: str | None = None,
        show_expression_as: ExpressionFormat | str = ExpressionFormat.TREE,
        show_filter_tokens: bool = False,
        show_advanced_properties: bool = False,
    ) -> 'IncrementalRefreshReport':
        """Create a new Incremental Refresh Report, using a data layout template
        from an existing Intelligent Cube.

        Args:
            connection: Strategy One connection object returned by
                `connection.Connection()`
            name (str, optional): Name of the Incremental Refresh Report
            destination_folder (Folder, str): Folder object or folder ID where
                the new object will be saved
            destination_folder_path (str, optional): Path of the folder in which
                to create the report. If both folder and path parameters are
                provided, `destination_folder` is used.
                The path has to be provided in the following format:
                    if it's inside of a project, example:
                        /MicroStrategy Tutorial/Public Objects
                    if it's a root folder, example:
                        /CASTOR_SERVER_CONFIGURATION/Users
            target_cube (OlapCube, SchemaObjectReference, dict, optional):
                Reference to an Intelligent Cube
            refresh_type (RefreshType, str, optional): Mode the report will be
                refreshed in. Accepted values: 'update', 'insert', 'delete',
                'update_only', 'replace'
            filter (Expression, dict, optional): Filtering expression for the
                report. Accepted values: 'tree', 'tokens'
            advanced_properties (dict, optional): Advanced properties of the
                report
            description (str, optional): Description of the MSTR object
            primary_locale (str, optional): The primary locale of the object,
                in the IETF BCP 47 language tag format, such as "en-US"
            show_expression_as (ExpressionFormat or str, optional): specify how
                expressions should be presented
                Available values:
                - None
                (expression is returned in "text" format)
                - `ExpressionFormat.TREE` or `tree`
                (expression is returned in `text` and `tree` formats)
                - `ExpressionFormat.TOKENS` or `tokens`
                (expression is returned in `text` and `tokens` formats)
            show_filter_tokens (bool, optional): Specify whether "qualification"
                is returned in "tokens" format,
                along with `text` and `tree` formats.
                - If omitted or false, only `text` and `tree`
                formats are returned.
                - If true, all `text`, "tree" and `tokens` formats are returned.
        """
        parameters = filter_params_for_func(cls.create_from_cube, locals(), ['cls'])

        return cls.create(**parameters, increment_type=cls.IncrementType.REPORT)

    def alter(
        self,
        name: str | None = None,
        comments: str | None = None,
        owner: str | User | None = None,
        target_cube: OlapCube | SchemaObjectReference | dict | None = None,
        increment_type: IncrementType | str | None = None,
        refresh_type: RefreshType | str | None = None,
        template: Template | dict | None = None,
        filter: Expression | dict | None = None,
        advanced_properties: dict | None = None,
        description: str | None = None,
        primary_locale: str | None = None,
    ):
        """Alter the Incremental Refresh Report object's properties.

        Args:
            name (str, optional): Name of the Incremental Refresh Report.
            comments (str, optional): Long description of the report
            owner: (str, User, optional): owner of the report
            target_cube (OlapCube, SchemaObjectReference, dict, optional):
                Reference to an Intelligent Cube
            increment_type (IncrementType, str, optional): Mode the report will
                be executed in. Accepted values: 'report', 'filter'
            refresh_type (RefreshType, str, optional): Mode the report will be
                refreshed in. Accepted values: 'update', 'insert', 'delete',
                'update_only', 'replace'
            template (Template, dict, optional): Template of information layout
                across dimensions
            filter (Expression, dict, optional): Filtering expression for the
                report. Accepted values: 'tree', 'tokens'
            advanced_properties (dict, optional): Advanced properties of the
                report
            description (str, optional): Description of the MSTR object
            primary_locale (str, optional): The primary locale of the object,
                in the IETF BCP 47 language tag format, such as "en-US"
        """
        if isinstance(owner, User):
            owner = owner.id
        if isinstance(target_cube, OlapCube):
            target_cube = SchemaObjectReference(
                object_id=target_cube.id,
                name=target_cube.name,
                sub_type=ObjectSubType.REPORT_CUBE,
            )
        filter = {} if self.filter is None and filter is None else filter  # NOSONAR
        arguments = filter_params_for_func(self.alter, locals(), exclude=['self'])

        self._alter_properties(**arguments)

    def change_increment_type_to_filter(self) -> None:
        """Change the Incremental Refresh Report's type to 'Filter'.
        This will reset the template to default.
        """
        self.fetch()
        self.template = None
        self.alter(increment_type=IncrementalRefreshReport.IncrementType.FILTER)

    def reset_template_to_default(self) -> None:
        """Reset the Incremental Refresh Report's template to default."""
        if self.increment_type == IncrementalRefreshReport.IncrementType.REPORT:
            self.change_increment_type_to_filter()
            self.alter(increment_type=IncrementalRefreshReport.IncrementType.REPORT)

    def get_preview_data(
        self,
        offset: int | None = None,
        limit: int | None = None,
        fields: str | None = None,
    ) -> pd.DataFrame:
        """Get preview data for the Incremental Refresh Report.

        Args:
            offset (int): Starting point within the collection of returned
                results. Used to control paging behavior. Default is 0.
            limit (int): limit the number of elements returned.
                If `None` (default), all objects are returned.
            fields (str, optional): A whitelist of top-level fields separated by
                commas.
        """
        response = refresh_api.create_incremental_refresh_report_instance(
            self.connection, self.id
        )
        instance_id = response.headers['X-MSTR-MS-Instance']
        refresh_api.request_incremental_refresh_report_preview_data(
            connection=self.connection,
            id=self.id,
            instance_id=instance_id,
            project_id=self.connection.project_id,
            fields=fields,
        )
        json = refresh_api.get_incremental_refresh_report_preview_data(
            connection=self.connection,
            id=self.id,
            instance_id=instance_id,
            project_id=self.connection.project_id,
            offset=offset,
            limit=limit,
            fields=fields,
        ).json()

        refresh_api.delete_incremental_refresh_report_instance(
            connection=self.connection,
            id=self.id,
            instance_id=instance_id,
        )

        parser = Parser(response=json, parse_cube=False)
        parser.parse(response=json)

        return parser.dataframe

    @property
    def prompts(self) -> dict:
        """Prompts of the report."""
        if self._prompts is None:
            prompts = reports_api.get_report_prompts(
                connection=self.connection, report_id=self.id
            ).json()
            self._prompts = [
                Prompt.from_dict(source=prompt, connection=self.connection)
                for prompt in prompts
            ]
        return self._prompts



from mstrio.connection import get_connection
from mstrio.project_objects.content_group import ContentGroup
from mstrio.project_objects.applications import Application
from mstrio.project_objects.dashboard import Dashboard
from mstrio.users_and_groups import User, UserGroup

conn = get_connection(workstationData, 'MicroStrategy Tutorial')

# Create a Content Group
cg = ContentGroup.create(
    connection=conn,
    name='Content Group for Application',
    color='#ffe4e1',
    opacity=100,
    email_enabled=True,
    recipients=[
        User(connection=conn, id='0870757440D0FEB7CDC449AAE6C18AFF'),
        UserGroup(connection=conn, id='E96685CD4E60068559F7DFAC7C2AA851'),
    ],
)

# Add some contents to the Content Group
# Add Contents to a Content Group
cg.update_contents(
    content_to_add=[
        Dashboard(connection=conn, id='F6252B9211E7B348312C0080EF55DB6A'),
        Dashboard(connection=conn, id='27BB740E11E7A7B40A650080EF856B88'),
    ]
)

# Create an Application
# To limit the Application to only the specified group, pass the
# ContentGroup ID in the content_bundle_ids of the home_library settings
# And make sure to set show_all_contents to False

# Only the necessary variables are defined here
app = Application.create(
    connection=conn,
    name='Application Workflow Showcase',
    description='This is a demo application created for content group.',
    home_screen=Application.HomeSettings(
        mode=0,
        home_library=Application.HomeSettings.HomeLibrary(
            content_bundle_ids=[cg.id],
            show_all_contents=False,
        ),
    ),
)



from typing import TYPE_CHECKING

from mstrio.types import ObjectSubTypes
from mstrio.utils.helper import deprecation_warning, find_object_with_name

from .agents import Agent, _list_implementation

if TYPE_CHECKING:
    from mstrio.connection import Connection
    from mstrio.server.project import Project


DEPRECATION_VERSION = "11.6.1.101"  # NOSONAR


def list_bots(
    connection: 'Connection',
    name: str | None = None,
    to_dictionary: bool = False,
    limit: int | None = None,
    project: 'Project | str | None' = None,
    project_id: str | None = None,
    project_name: str | None = None,
    **filters,
) -> list['Bot'] | list[dict]:
    """Get a list of Bots.

    Args:
        connection (Connection): Strategy One connection object returned
            by 'connection.Connection()'
        name (str, optional): characters that the dashboard name must contain
        to_dictionary (bool, optional): if True, return Bots as a
            list of dicts
        limit (int, optional): limit the number of elements returned.
            If `None` (default), all objects are returned.
        project (Project | str, optional): Project object or ID or name
            specifying the project. May be used instead of `project_id` or
            `project_name`.
        project_id (str, optional): Project ID
        project_name (str, optional): Project name
        **filters: Available filter parameters: ['name', 'id', 'type',
            'subtype', 'date_created', 'date_modified', 'version',
            'owner', 'ext_type', 'view_media', 'certified_info']

    Returns:
        A list of Bots objects or dictionaries.
    """
    deprecation_warning('list_bots', None, DEPRECATION_VERSION, module=False)

    return _list_implementation(
        Bot,
        connection=connection,
        name=name,
        to_dictionary=to_dictionary,
        limit=limit,
        project=project,
        project_id=project_id,
        project_name=project_name,
        **filters,
    )


class Bot(Agent):
    _OBJECT_SUBTYPES = [ObjectSubTypes.DOCUMENT_BOT] + Agent._OBJECT_SUBTYPES

    def __init__(
        self, connection: 'Connection', name: str | None = None, id: str | None = None
    ) -> None:
        if id is None:
            if name is None:
                raise ValueError("Please specify either 'name' or 'id'.")

            bot = find_object_with_name(
                connection=connection,
                cls=self.__class__,
                name=name,
                listing_function=list_bots,
            )
            id = bot['id']
        super().__init__(
            connection=connection,
            id=id,
            name=name,
        )

        deprecation_warning(
            'mstrio.project_objects.bots',
            (
                'mstrio.project_objects.agents'
                if self._subtype in Agent._get_subtypes_as_raw_values()
                else None
            ),
            DEPRECATION_VERSION,
        )
